# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Classes for arranging the input and output for nodes.

Exported classes:

Input -- An input for a node (holding datatype).
Output -- The output of a node (holding datatype and value).
ConstantOutput -- The output of a node (holding datatype and value).

.. warning::
   Don't mess with the Link, Input and Output internals from other places.
   There will be a huge chances of breaking the network functionality!
"""
from abc import abstractmethod, abstractproperty
from collections import OrderedDict
from typing import Union

import sympy

from .. import exceptions
from ..abc.serializable import Serializable
from ..abc.updateable import Updateable
from ..core.cardinality import create_cardinality
from ..core.dimension import HasDimensions, Dimension
from ..core.interface import InputSpec, OutputSpec
from ..datatypes import DataType, types
from ..helpers import log
from ..utils.dicteq import dicteq


class BaseInputOutput(HasDimensions, Updateable, Serializable):

    """
    Base class for Input and Output classes. It mainly implements the
    properties to access the data from the underlying ParameterDescription.
    """
    description_type = None

    def __init__(self, node, description):
        """Instantiate a BaseInputOutput

        :param node: the parent node the input/output belongs to.
        :param description: the :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
               describing the input/output.
        :return: created BaseInputOutput
        :raises FastrTypeError: if description is not of class
                                :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
        :raises FastrDataTypeNotAvailableError: if the DataType requested cannot be found in the ``types``
        """
        super(BaseInputOutput, self).__init__()

        self._node = node

        # Get DataType
        if description.datatype in types:
            self._description = description
            self._datatype = types[description.datatype]

            # Create a validator for the cardinality
            self.cardinality_spec = create_cardinality(description.cardinality, self)
        else:
            raise exceptions.FastrDataTypeNotAvailableError('DataType {} does not exist'.format(description.datatype))

    def __ne__(self, other):
        """
        Check two Node instances for inequality. This is the inverse of __eq__

        :param other: the other instances to compare to
        :type other: BaseInputOutput
        :returns: True if unequal, False otherwise
        """
        if not isinstance(self, type(other)):
            return NotImplemented

        return not self == other

    def __iter__(self):
        """
        This function is blocked to avoid support for iteration using a lecacy __getitem__ method.

        :return: None
        :raises FastrNotImplementedError: always
        """
        raise exceptions.FastrNotImplementedError('Not iterable, this function is to block legacy iteration using getitem')

    def __getstate__(self):
        """
        Retrieve the state of the BaseInputOutput

        :return: the state of the object
        :rtype dict:
        """
        state = super(BaseInputOutput, self).__getstate__()
        state['id'] = self.id
        state['datatype'] = self.datatype.id
        return state

    def __setstate__(self, state):
        """
        Set the state of the BaseInputOutput by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(BaseInputOutput, self).__setstate__(state)

        if 'description' in state:
            description = state.pop('description')
            if description['datatype'] in types:
                # Reconstruct description object
                description_type = self.description_type
                args = [description[x] for x in description_type._fields]
                self._description = description_type(*args)

                self._datatype = types[description['datatype']]

                # Create a validator for the cardinality
                self.cardinality_spec = create_cardinality(self._description.cardinality, self)
            else:
                raise exceptions.FastrDataTypeNotAvailableError('DataType {} does not exist'.format(description.datatype))

        if 'datatype' in state:
            self._datatype = types[state['datatype']]


    def __repr__(self):
        """
        Get a string representation for the Input/Output

        :return: the string representation
        :rtype: str
        """
        return f'<{type(self).__name__}: {self.fullid}>'

    @property
    def datatype(self):
        """
        The datatype of this Input/Output
        """
        return self._datatype

    @datatype.setter
    def datatype(self, value):
        """
        The datatype of this Input/Output (setter)
        """
        self._datatype = value

    @property
    def description(self):
        """
        The description object of this input/output
        """
        return self._description

    def cardinality(self, key=None, job_data=None):
        """
        Determine the cardinality of this Input/Output. Optionally a key can be
        given to determine for a sample.

        :param key: key for a specific sample
        :return: the cardinality
        :rtype: int, sympy.Symbol, or None
        """

        # We need to key for the signature in subclasses, shut pylint up
        # pylint: disable=unused-argument,no-self-use

        raise exceptions.FastrNotImplementedError('Purposefully not implemented')

    @property
    def id(self):
        """
        Id of the Input/Output
        """
        return self._description.id

    @property
    def node(self):
        """
        The NodeRun to which this Input/Output belongs
        """
        return self._node

    @property
    def required(self):
        """
        Flag indicating that the Input/Output is required
        """
        return self._description.required

    @abstractproperty
    def fullid(self):
        """
        The fullid of the Input/Output, the fullid should be unnique and
        makes the object retrievable by the network.
        """
        raise exceptions.FastrNotImplementedError('Purposefully not implemented')

    def check_cardinality(self, key=None, planning=False) -> bool:
        """
        Check if the actual cardinality matches the cardinality specified in
        the ParameterDescription. Optionally you can use a key to test for a
        specific sample.

        :param key: sample_index (tuple of int) or
                    :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
                    for desired sample
        :return: flag indicating that the cardinality is correct
        :rtype: bool
        :raises FastrCardinalityError: if the Input/Output has an incorrect
                cardinality description.
        """
        return self.cardinality_spec.validate(
            payload=None,
            cardinality=self.cardinality(key)
        )


class BaseInput(BaseInputOutput):
    """
    Base class for all inputs.
    """
    description_type = InputSpec

    def __init__(self, node, description):
        """
        Instantiate a BaseInput

        :param node: the parent node the input/output belongs to.
        :param description: the :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
                            describing the input/output.
        :return: the created BaseInput
        :raises FastrTypeError: if description is not of class
                                :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
        :raises FastrDataTypeNotAvailableError: if the DataType requested cannot be found in the ``types``
        """
        if not isinstance(description, InputSpec):
            log.error('Description has type "{}" (must be ParameterDescription)'.format(type(description).__name__))
            raise exceptions.FastrTypeError('An input must be constructed based on an '
                                            'object of a class derived from NodeRun and an '
                                            'object of class InputSpec')

        super(BaseInput, self).__init__(node, description)

    @abstractmethod
    def itersubinputs(self):
        """
        Iterator over the SubInputs

        :return: iterator

        example:

        .. code-block:: python

          >>> for subinput in input_a.itersubinputs():
                  print subinput

        """
        raise exceptions.FastrNotImplementedError('Purposefully not implemented')

    def __lshift__(self, other):
        if not isinstance(other, (BaseOutput, list, tuple, dict, OrderedDict)):
            return NotImplemented

        return self.create_link_from(other)

    def __rrshift__(self, other):
        if not isinstance(other, (BaseOutput, list, tuple, dict, OrderedDict)):
            return NotImplemented

        return self.create_link_from(other)

    @property
    def default(self):
        """
        Default value
        """
        return self._description.default

    @property
    def item_index(self):
        return None

    def constant_id(self) -> str:
        """
        The id that should be used for a constant created to serve this input.
        """
        return f'const__{self.node.id}__{self.id}'

    def create_link_from(self, value):
        if isinstance(value, BaseOutput):
            if self.node.parent is not value.node.parent:
                message = 'Cannot create links between members of different Network'
                log.warning(message)

            network = value.node.parent
            if network is None:
                message = 'Cannot create links between non-network-attached Nodes'
                log.warning(message)
            else:
                log.debug('Linking {} to {}'.format(value.fullid, self.fullid))
                return network.create_link(value, self)
        elif isinstance(value, (list, dict, OrderedDict)) or\
                (isinstance(value, tuple) and all(not isinstance(x, BaseOutput) for x in value)):
            # This is data for a ConstantNode, so create one and set it
            # First make sure the stepid of the new ConstantNode will match the stepid of the current Node
            inp = self
            for k, i in inp.node.parent.stepids.items():
                if inp.node in i:
                    stepid = k
                    break
            else:
                stepid = None
            network = inp.node.parent

            # Create constant node with correct type and id
            const_node = network.create_constant(datatype=inp.datatype,
                                                 data=value,
                                                 id_=self.constant_id,
                                                 stepid=stepid)
            return network.create_link(const_node.output, self)
        elif isinstance(value, tuple) and isinstance(self, Input):
            # First remove all current links
            self.clear()
            new_links = []

            # Create all required links, find all consecutive parts of non-outputs and
            # create ConstantNodes for those, link all outputs separately
            current_part = []
            for element in value:
                if isinstance(element, BaseOutput):
                    # If there were non-outputs found, first combine those
                    # before creating a link from the output
                    if len(current_part) > 0:
                        new_links.append(self.append(tuple(current_part)))
                        current_part = []
                    new_links.append(self.append(element))
                else:
                    current_part.append(element)
            if len(current_part) > 0:
                new_links.append(self.append(tuple(current_part)))

            # Return a tuple of all links created
            return tuple(new_links)
        else:
            # We assume that if above tests are all False, a constant is given.
            return self.create_link_from([(value,)])

    def check_cardinality(self, key=None, planning=False) -> bool:
        """
        Check if the actual cardinality matches the cardinality specified in
        the ParameterDescription. Optionally you can use a key to test for a
        specific sample.

        :param key: sample_index (tuple of int) or
                    :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
                    for desired sample
        :return: flag indicating that the cardinality is correct
        :rtype: bool
        :raises FastrCardinalityError: if the Input/Output has an incorrect
                cardinality description.
        """
        cardinality = self.cardinality(key)

        if not isinstance(cardinality, int):
            return planning

        if not self.required and isinstance(cardinality, int) and cardinality == 0:
            return True

        return self.cardinality_spec.validate(
            payload=None,
            cardinality=cardinality
        )


class Input(BaseInput):
    """
    Class representing an input of a node. Such an input will be connected
    to the output of another node or the output of an constant node to provide
    the input value.
    """

    def __init__(self, node, description):
        """
        Instantiate an input.

        :param node: the parent node of this input.
        :type node: :py:class:`NodeRun <fastr.planning.node.NodeRun>`
        :param ParameterDescription description: the ParameterDescription of the input.
        :return: the created Input
        """
        self._source = {}
        super(Input, self).__init__(node, description)
        self._input_group = 'default'

    def __eq__(self, other):
        """Compare two Input instances with each other. This function ignores
        the parent node and update status, but tests rest of the dict for equality.

        :param other: the other instances to compare to
        :type other: :py:class:`Input <fastr.planning.inputoutput.Input>`
        :returns: True if equal, False otherwise
        :rtype: bool
        """
        if not isinstance(other, type(self)):
            return NotImplemented

        dict_self = dict(vars(self))
        del dict_self['_node']
        del dict_self['_status']

        dict_other = dict(vars(other))
        del dict_other['_node']
        del dict_other['_status']

        return dicteq(dict_self, dict_other)

    def __getstate__(self):
        """
        Retrieve the state of the Input

        :return: the state of the object
        :rtype dict:
        """
        state = super(Input, self).__getstate__()
        state['input_group'] = self.input_group

        return state

    def __setstate__(self, state):
        """
        Set the state of the Input by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(Input, self).__setstate__(state)
        self._input_group = state['input_group']

    def __getitem__(self, key: Union[int, str]) -> 'Union[SubInput, NamedSubInput]':
        """
        Retrieve an item from this Input.

        :param key: the key of the requested item
        :return: The :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
                 corresponding with the key will be returned.
        :raises FastrTypeError: if key is not of a valid type
        :raises FastrKeyError: if the key is not found
        """
        if not isinstance(key, (int, str)):
            raise exceptions.FastrTypeError('Input indices must a int or str'
                                            ' not {}'.format(type(key).__name__))

        if key not in self.source:
            # This is to allow for linking against inputs['key'][0]
            try:
                key = int(key)
            except ValueError:
                pass  # No problem, just go for the str

            if isinstance(key, int):
                self.source[key] = SubInput(self)
            else:
                self.source[key] = NamedSubInput(self)

        return self.source[key]

    def __setitem__(self, key: Union[int, str], value):
        """
        Create a link between a SubInput of this Inputs and an Output/Constant

        :param key: the key of the SubInput
        :type key: int, str
        :param value: the target to link, can be an output or a value to create a constant for
        :type value: BaseOutput, list, tuple, dict, OrderedDict
        :raises FastrTypeError: if key is not of a valid type
        """
        if not isinstance(key, (int, str)):
            raise exceptions.FastrTypeError('The key of an SubInput to set should be an '
                                            'int or str (found {})'.format(type(key).__name__))

        if key not in self.source:
            if isinstance(key, int):
                subin = SubInput(self)
            else:
                subin = NamedSubInput(self)
            self.source[key] = subin

        self.source[key].create_link_from(value)

    def __str__(self):
        """
        Get a string version for the Input

        :return: the string version
        :rtype: str
        """
        return f'<Input: {self.fullid})>'

    def cardinality(self, key=None, job_data=None):
        """
        Cardinality for an Input is the sum the cardinalities of the SubInputs,
        unless defined otherwise.

        :param key: key for a specific sample, can be sample index or id
        :type key: tuple of int or :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the cardinality
        :rtype: int, sympy.Symbol, or None
        """

        cardinality = 0

        for subinput in self.source.values():
            cardinality += subinput.cardinality(key, job_data)

        return cardinality

    @property
    def datatype(self):
        """
        The datatype of this Input
        """
        return self._datatype

    @datatype.setter
    def datatype(self, value):
        # This does not differ, as it is a property
        # pylint: disable=arguments-differ
        self._datatype = value
        for subinput in self.itersubinputs():
            subinput.datatype = value

    @property
    def dimensions(self):
        """
        The list names of the dimensions in this Input. This will be a list of str.
        """
        subinputs = list(self.itersubinputs())
        sizes = [sub.size for sub in subinputs]

        unique_sizes = set(sizes) - {(0,), (1,), ()}

        if len(unique_sizes) > 1:
            nr_non_symbolic_sizes = sum(all(not isinstance(x, sympy.Symbol) for x in size) for size in unique_sizes)

            if nr_non_symbolic_sizes == 0:
                max_dimensions = max(len(x) for x in unique_sizes)
                for subinput in subinputs:
                    if len(subinput.size) == max_dimensions and subinput.size not in ((0,), (1,), ()):
                        return subinput.dimensions

            raise exceptions.FastrSizeMismatchError('Cannot determine dimensions: sizes of SubInputs do not match!')
        elif len(unique_sizes) == 1:
            return subinputs[sizes.index(unique_sizes.pop())].dimensions
        elif (1,) in sizes:
            return subinputs[sizes.index((1,))].dimensions
        elif (0,) in sizes:
            return subinputs[sizes.index((0,))].dimensions
        else:
            return []

    @property
    def id(self):
        """
        Id of the Input/Output
        """
        return self._description.id

    @property
    def fullid(self) -> str:
        """
        The full defining ID for the Input
        """
        if self.node is not None:
            return '{}/inputs/{}'.format(self.node.fullid, self.id)
        else:
            return 'fastr://ORPHANED/inputs/{}'.format(self.id)

    @property
    def constant_id(self) -> str:
        """
        The id for a constant node that is attached to this input.
        """
        return f'const__{self.node.id}__{self.id}'

    @property
    def input_group(self) -> str:
        """
        The id of the :py:class:`InputGroup <fastr.planning.node.InputGroup>` this
        Input belongs to.
        """
        return self._input_group

    @input_group.setter
    def input_group(self, value: str):
        """
        The id of the :py:class:`InputGroup <fastr.planning.node.InputGroup>` this
        Input belongs to. (setter)
        """
        self._input_group = value
        self.node.update()

    @property
    def source(self):
        """
        The mapping of :py:class:`SubInputs <fastr.planning.inputoutput.SubInput>`
        that are connected and have more than 0 elements.
        """
        return self._source

    @source.setter
    def source(self, value):
        """
        The list of :py:class:`SubInputs <fastr.planning.inputoutput.SubInput>`
        that are connected and have more than 0 elements. (setter)
        """
        self.clear()

        self._source = {0: SubInput(self)}
        self._source[0].source = value

    def get_sourced_nodes(self):
        """
        Get a list of all :py:class:`Nodes <fastr.planning.node.Node>` connected as sources to this Input

        :return: list of all connected :py:class:`Nodes <fastr.planning.node.Node>`
        :rtype: list
        """
        sourced_nodes = []
        for subinput in self.itersubinputs():
            for node in subinput.get_sourced_nodes():
                if node not in sourced_nodes:
                    sourced_nodes.append(node)
        return sourced_nodes

    def get_sourced_outputs(self):
        """
        Get a list of all :py:class:`Outputs <fastr.planning.inputoutput.Output>` connected as sources to this Input

        :return: tuple of all connected :py:class:`Outputs <fastr.planning.inputoutput.Output>`
        :rtype: tuple
        """
        sourced_outputs = []
        for subinput in self.itersubinputs():
            for output in subinput.get_sourced_outputs():
                if output not in sourced_outputs:
                    sourced_outputs.append(output)
        return tuple(sourced_outputs)

    def index(self, value):
        """
        Find index of a SubInput

        :param value: the :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
                      to find the index of
        :type value: :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
        :return: key
        :rtype: int, str
        """
        for key, val in self.source.items():
            if val is value:
                return key
        else:
            return None

    def remove(self, value: 'Union[SubInput, NamedSubInput, Link]'):
        """
        Remove a SubInput from the SubInputs list based on the connected Link.

        :param value: the :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
                      or :py:class:`Link <fastr.planning.link.Link>`
                      to removed from this Input
        :type value: :py:class:`SubInput <fastr.planning.link.Link>`, <fastr.planning.inputoutput.SubInput>`
        """
        to_remove = []
        # Find keys to remove
        for key, subinput in self.source.items():
            if subinput is value or (len(subinput.source) == 1 and subinput.source[0] is value):
                to_remove.append(key)

        # Remove actual data
        for key in to_remove:
            target = self.source.pop(key)

            if target.source is not None:
                for link in target.source:
                    link.destroy()

    def clear(self):
        for key in self.source.keys():
            subinput = self.source.pop(key)
            if subinput.source is not None:
                subinput.source[0].destroy()

    def insert(self, index):
        """
        Insert a new SubInput at index in the sources list

        :param int key: positive integer for position in _source list to insert to
        :return: newly inserted :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
        :rtype: :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
        """
        newsub = SubInput(self)
        self.source[index] = newsub
        return newsub

    def append(self, value):
        """
        When you want to append a link to an Input, you can use the append
        property. This will automatically create a new SubInput to link to.

        example:

        .. code-block:: python

          >>> link = node2['input'].append(node1['output'])

        will create a new SubInput in node2['input'] and link to that.
        """
        new_sub = SubInput(self)
        # Get the next index-like key to use
        new_key = max([-1] + [x for x in self.source.keys() if isinstance(x, int)]) + 1
        self.source[new_key] = new_sub
        return new_sub.create_link_from(value)

    def itersubinputs(self):
        """
        Iterate over the :py:class:`SubInputs <fastr.planning.inputoutput.SubInput>`
        in this Input.

        :return: iterator yielding  :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`


        example:

        .. code-block:: python

          >>> for subinput in input_a.itersubinputs():
                  print subinput

        """
        for subinput in self.source.values():
            yield subinput

    def _update(self, key, forward=True, backward=False):
        """Update the validity of the Input and propagate the update downstream.
        An Input is valid if:

        * All SubInputs are valid (see :py:meth:`SubInput.update <fastr.planning.inputoutput.SubInput.update>`)
        * Cardinality is correct
        * If Input is required, it must have a size larger than (0,)

        """
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.fullid))
        for subinput in self.itersubinputs():
            subinput.update(key, forward, backward)

        valid = True
        messages = []
        for subinput in self.itersubinputs():
            if not subinput.valid:
                valid = False
                for message in subinput.messages:
                    messages.append('SubInput {} is not valid: {}'.format(subinput.fullid, message))

        if not self.check_cardinality(planning=True):
            cardinality = self.cardinality()

            # If the cardinality is 0 and Input is not required, this is fine,
            # all other cases where the cardinality check fails are not allowed
            valid = False

            messages.append(('Input "{}" cardinality ({}) is not valid (must'
                             ' be {}, required is {})').format(self.id,
                                                               cardinality,
                                                               self._description.cardinality,
                                                               self.required))
            if isinstance(self._description.cardinality, str) and self._description.cardinality.startswith('as:'):
                target = self._description.cardinality[3:]
                if target in self.node.inputs:
                    target_cardinality = self.node.inputs[target].cardinality()
                    messages.append('Target input {} has cardinality {}'.format(target, target_cardinality))

        if self.size is None:
            valid = False
            messages.append('Cannot determine size of Input "{}"'.format(self.id))

        log.debug('Size: {}'.format(self.size))
        if self.required and (len([x for x in self.size if x != 0]) == 0):
            valid = False
            nodes = ', '.join([x.id for x in self.get_sourced_nodes()])
            messages.append(('Required Input "{}" cannot have size 0. Input obtained'
                             ' from nodes: {}').format(self.id, nodes))

        self._status['valid'] = valid
        self._status['messages'] = messages

        # Update downstream
        self.node.update(key, forward, backward)


class NamedSubInput(Input):
    """
    A named subinput for cases where the value of an input is mapping.
    """
    def __init__(self, parent):
        super().__init__(parent.node, parent.description)
        self.parent = parent

    def __getitem__(self, key: int) -> 'SubInput':
        """
        Retrieve an item (a SubInput) from this NamedSubInput.

        :param key: the key of the requested item
        :return: The :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
                 corresponding with the key will be returned.
        :raises FastrTypeError: if key is not of a valid type
        :raises FastrKeyError: if the key is not found
        """
        if not isinstance(key, int):
            raise exceptions.FastrTypeError('NamedSubInput indices must a int'
                                            ' not {}'.format(type(key).__name__))

        if key not in self.source:
            self.source[key] = SubInput(self)

        return self.source[key]

    def __str__(self):
        """
        Get a string version for the NamedSubInput

        :return: the string version
        :rtype: str
        """
        return f'<NamedSubInput: {self.fullid})>'

    @property
    def fullid(self):
        """
        The full defining ID for the SubInput
        """
        return '{}/{}'.format(self.parent.fullid, self.parent.index(self))

    @property
    def constant_id(self) -> str:
        """
        The id for a constant node that is attached to this input.
        """
        return f'{self.parent.constant_id}__{self.item_index}'

    @property
    def item_index(self):
        return self.parent.index(self)


class SubInput(BaseInput):
    """
    This class is used by :py:class:`Input <fastr.planning.inputoutput.Input>` to
    allow for multiple links to an :py:class:`Input <fastr.planning.inputoutput.Input>`.
    The SubInput class can hold only a single Link to a (Sub)Output, but behaves
    very similar to an :py:class:`Input <fastr.planning.inputoutput.Input>` otherwise.
    """

    def __init__(self, input_):
        """
        Instantiate an SubInput.

        :param input_: the parent of this SubInput.
        :type input_: :py:class:`Input <fastr.planning.inputoutput.Input>`
        :return: the created SubInput
        """
        self._source = None

        if not isinstance(input_, Input):
            raise exceptions.FastrTypeError('First argument for a SubInput constructor should be an Input')

        self.parent = input_
        super(SubInput, self).__init__(self.node, self.description)

        self.datatype = input_.datatype
        if self.parent.valid:
            self.update()

    def __getitem__(self, key):
        """
        Retrieve an item from this SubInput.

        :param key: the index of the requested item
        :type key: int
        :return: the corresponding :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
        :rtype: :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
        :raises FastrTypeError: if key is not of a valid type

        .. note:: As a SubInput has only one SubInput, only requesting int key
                  0 or -1 is allowed, and it will return self
        """

        if not isinstance(key, int):
            raise exceptions.FastrTypeError('SubInput indices must be an int, not {}'.format(type(key).__name__))

        if not -1 <= key < 1:
            raise exceptions.FastrIndexError('SubInput index out of range (key: {})'.format(key))

        return self

    def __eq__(self, other):
        """Compare two SubInput instances with each other. This function ignores
        the parent, node, source and update status, but tests rest of the dict
        for equality.

        :param other: the other instances to compare to
        :type other: SubInput
        :returns: True if equal, False otherwise
        """
        if not isinstance(other, type(self)):
            return NotImplemented

        dict_self = dict(vars(self))
        del dict_self['_node']
        del dict_self['parent']
        del dict_self['_source']
        del dict_self['_status']

        dict_other = dict(vars(other))
        del dict_other['_node']
        del dict_other['parent']
        del dict_other['_source']
        del dict_other['_status']

        return dicteq(dict_self, dict_other)

    def __getstate__(self):
        """
        Retrieve the state of the SubInput

        :return: the state of the object
        :rtype dict:
        """
        state = super(SubInput, self).__getstate__()
        return state

    def __setstate__(self, state):
        """
        Set the state of the SubInput by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(SubInput, self).__setstate__(state)

        if not hasattr(self, '_source'):
            self._source = None

    def __str__(self):
        """
        Get a string version for the SubInput

        :return: the string version
        :rtype: str
        """
        return f'<SubInput: {self.fullid}'

    def cardinality(self, key=None, job_data=None):
        """
        Get the cardinality for this SubInput. The cardinality for a SubInputs
        is defined by the incoming link.

        :param key: key for a specific sample, can be sample index or id
        :type key: :py:class:`SampleIndex <fastr.core.sampleidlist.SampleIndex>` or :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the cardinality
        :rtype: int, sympy.Symbol, or None
        """
        if self.source is not None:
            return self.source[0].cardinality(index=key)
        else:
            return 0

    @property
    def description(self):
        return self.parent.description

    @property
    def dimensions(self):
        """
        List of dimension for this SubInput
        """
        return self.source[0].dimensions

    @property
    def fullid(self):
        """
        The full defining ID for the SubInput
        """
        return '{}/{}'.format(self.parent.fullid, self.parent.index(self))

    @property
    def constant_id(self) -> str:
        """
        The id for a constant node that is attached to this input.
        """
        return f'{self.parent.constant_id}__{self.item_index}'

    @property
    def item_index(self):
        index = self.parent.index(self)
        return index

    @property
    def input_group(self):
        """
        The id of the :py:class:`InputGroup <fastr.planning.node.InputGroup>` this
        SubInputs parent belongs to.
        """
        return self.parent.input_group

    @property
    def node(self):
        """
        The Node to which this SubInputs parent belongs
        """
        return self.parent.node

    @property
    def source_output(self):
        """
        The :py:class:`Output <fastr.planning.inputoutput.Output>`
        linked to this SubInput
        """
        if self.source is not None and len(self.source) > 0:
            return self.source[0].source
        else:
            return None

    @property
    def source(self):
        """
        A list with the source :py:class:`Link <fastr.planning.link.Link>`.
        The list is to be compatible with :py:class:`Input <fastr.planning.inputoutput.Input>`
        """
        if self._source is None:
            self.parent.remove(self)
            return []

        return [self._source]

    @source.setter
    def source(self, value):
        """
        Set new source, make sure previous link to source is released
        """
        if value is self._source:
            return

        if self._source is not None:
            self._source.destroy()

        if value is None:
            self.parent.remove(self)

        self._source = value

    def get_sourced_nodes(self):
        """
        Get a list of all :py:class:`Nodes <fastr.planning.node.Node>` connected as sources to this SubInput

        :return: list of all connected :py:class:`Nodes <fastr.planning.node.Node>`
        :rtype: list
        """
        return [x.source.node for x in self.source]

    def get_sourced_outputs(self):
        """
        Get a list of all :py:class:`Outputs <fastr.planning.inputoutput.Output>` connected as sources to this SubInput

        :return: list of all connected :py:class:`Outputs <fastr.planning.inputoutput.Output>`
        :rtype: list
        """
        return [x.source for x in self.source]

    def remove(self, value):
        """
        Remove a SubInput from parent Input.

        :param value: the :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
                      to removed from this Input
        :type value: :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`
        """
        # Pass on to the parent Input
        self.parent.remove(value)

    def _update(self, key, forward=True, backward=False):
        """Update the validity of the SubInput and propagate the update downstream.
        A SubInput is valid if:

        * the source Link is set and valid (see :py:meth:`Link.update <fastr.planning.link.Link.update>`)

        """
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.fullid))
        valid = True
        messages = []
        if len(self.source) == 0:
            self.parent.remove(self)
            valid = False
            messages.append('No source in this SubInput, removing!')
        elif not self.source[0].valid:
            valid = False
            messages.append('SubInput source ({}) is not valid'.format(self.source[0].id))
            messages.extend(self.source[0].messages)

        self._status['valid'] = valid
        self._status['messages'] = messages

        # Update downstream
        self.parent.update(key, forward, backward)

    def iteritems(self):
        """
        Iterate over the :py:class:`SampleItems <fastr.core.sampleidlist.SampleItem>`
        that are in the SubInput.

        :return: iterator yielding :py:class:`SampleItem <fastr.core.sampleidlist.SampleItem>` objects
        """
        for item in self.source.items():
            yield item

    def itersubinputs(self):
        """Iterate over SubInputs (for a SubInput it will yield self and stop iterating after that)

        :return: iterator yielding  :py:class:`SubInput <fastr.planning.inputoutput.SubInput>`

        example:

        .. code-block:: python

          >>> for subinput in input_a.itersubinputs():
                  print subinput

        """
        yield self


class MacroInput(Input):
    @property
    def input_group(self):
        return self._input_group

    @input_group.setter
    def input_group(self, value):
        raise exceptions.FastrNotImplementedError("Input groups are not settable for MacroNodes"
                                                  " (all inputs have to be in the default group)")


class BaseOutput(BaseInputOutput):
    """
    Base class for all outputs.
    """
    description_type = OutputSpec

    def __init__(self, node, description):
        """Instantiate a BaseOutput

        :param node: the parent node the output belongs to.
        :param description: the :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
               describing the output.
        :return: created BaseOutput
        :raises FastrTypeError: if description is not of class
                                :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
        :raises FastrDataTypeNotAvailableError: if the DataType requested cannot be found in the ``types``
        """
        if not isinstance(description, self.description_type):
            log.error('Description has type "{}" (must be ParameterDescription)'.format(type(description).__name__))
            raise exceptions.FastrTypeError('An output must be constructed based on an '
                                            'object of a class derived from Node and an '
                                            'object of class OutputSpec')
        super().__init__(node, description)

    @property
    def blocking(self):
        """
        Flag indicating that this Output will cause blocking in the execution
        """
        return not self.cardinality_spec.predefined

    @property
    def automatic(self):
        """
        Flag indicating that the Output is generated automatically
        without being specified on the command line
        """
        return self._description.automatic


class Output(BaseOutput):
    """
    Class representing an output of a node. It holds the output values of
    the tool ran. Output fields can be connected to inputs of other nodes.
    """

    def __init__(self, node, description):
        """Instantiate an Output

        :param node: the parent node the output belongs to.
        :param description: the :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
               describing the output.
        :return: created Output
        :raises FastrTypeError: if description is not of class
                                :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
        :raises FastrDataTypeNotAvailableError: if the DataType requested cannot be found in the ``types``
        """
        self._suboutputlist = {}
        self._samples = None
        super(Output, self).__init__(node, description)
        self._listeners = []
        self._preferred_types = []

    def __str__(self):
        """
        Get a string version for the Output

        :return: the string version
        :rtype: str
        """
        return '<Output: {}>'.format(self.fullid)

    def __eq__(self, other):
        """
        Compare two Output instances with each other. This function ignores
        the parent node, listeners and update status, but tests rest of the
        dict for equality.

        :param other: the other instances to compare to
        :type other: fastr.planning.inputoutput.Output
        :returns: True if equal, False otherwise
        :rtype: bool
        """
        if not isinstance(other, type(self)):
            return NotImplemented

        dict_self = dict(vars(self))
        del dict_self['_node']
        del dict_self['_listeners']
        del dict_self['_status']

        dict_other = dict(vars(other))
        del dict_other['_node']
        del dict_other['_listeners']
        del dict_other['_status']

        return dicteq(dict_self, dict_other)

    def __getitem__(self, key: Union[int, slice]) -> 'SubOutput':
        """
        Retrieve an item from this Output. The returned value depends on what type of key used:

        * Retrieving data using index tuple: [index_tuple]
        * Retrieving data sample_id str: [SampleId]
        * Retrieving a list of data using SampleId list: [sample_id1, ..., sample_idN]
        * Retrieving a :py:class:`SubOutput <fastr.planning.inputoutput.SubOutput>` using an int or slice: [n] or [n:m]

        :param key: the key of the requested suboutput, can be a numberor slice
        :return: the :py:class:`SubOutput <fastr.planning.inputoutput.SubOutput>`
                 for the corresponding index
        :raises FastrTypeError: if key is not of a valid type
        """

        if isinstance(key, (int, slice)):
            return self._get_suboutput(key)
        else:
            raise exceptions.FastrTypeError(
                'Index of SubOutput should be int/slice, found {}'.format(type(key).__name__)
            )

    def __getstate__(self):
        """
        Retrieve the state of the Output

        :return: the state of the object
        :rtype dict:
        """
        state = super(Output, self).__getstate__()

        # Add specific fields to the state
        state['suboutputs'] = [x.__getstate__() for x in self._suboutputlist.values()]
        if self._preferred_types is not None:
            state['preferred_types'] = [x.id for x in self._preferred_types]
        else:
            state['preferred_types'] = None

        return state

    def __setstate__(self, state):
        """
        Set the state of the Output by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(Output, self).__setstate__(state)

        if state['preferred_types'] is not None:
            self._preferred_types = [types[x] for x in state['preferred_types']]
        else:
            self._preferred_types = None

        suboutputlist = []
        for substate in state['suboutputs']:
            suboutput = SubOutput(self, slice(None))
            suboutput.__setstate__(substate)
            suboutputlist.append((suboutput.indexrep, suboutput))

        # Re-create the dict from the array
        self._suboutputlist = dict(suboutputlist)
        self._listeners = []

    def _cast_to_storetype(self, value):
        """
        Cast a given value to a DataType that matches this Outputs datatype.

        :param value: value to cast
        :return: cast value
        :rtype: DataType matching self.datatype
        """
        if isinstance(value, self.datatype):
            return value

        log.info('CAST VALUE: [{}] {!r} / {}'.format(type(value).__name__,
                                                           value,
                                                           value))

        storetype = types.match_types(self.datatype, type(value))

        if storetype is None:
            storetype = types.match_types(self.datatype)

        if not isinstance(value, storetype):
            if isinstance(value, DataType):
                log.warning('Changing value type from {} to {}'.format(type(value), storetype))
            value = storetype(str(value))

        return value

    def _get_suboutput(self, key):
        """
        Get a suboutput based on the key

        :param int, slice key: The key of the suboutput
        :return: the suboutput
        """
        # Get a string representation of the key
        if isinstance(key, slice):
            keystr = '{}:{}'.format(key.start, key.stop)
            keystr = keystr.replace('None', '')

            if key.step is not None and key.step != 1:
                keystr = '{}:{}'.format(keystr, key.step)
        else:
            keystr = str(key)

        if keystr in self._suboutputlist:
            # Re-use the same SubOutput
            subout = self._suboutputlist[keystr]
        else:
            # Create the desired SubOutput object
            subout = SubOutput(self, key)
            self._suboutputlist[keystr] = subout

        return subout

    def cardinality(self):
        """
        Cardinality of this Output, may depend on the inputs of the parent Node.

        :return: the cardinality
        :rtype: int, sympy.Symbol, or None
        :raises FastrCardinalityError: if cardinality references an invalid :py:class:`Input <fastr.planning.inputoutput.Input>`
        :raises FastrTypeError: if the referenced cardinality values type cannot be case to int
        :raises FastrValueError: if the referenced cardinality value cannot be case to int
        """
        return self.cardinality_spec.calculate_planning_cardinality()

    @property
    def datatype(self):
        """
        The datatype of this Output
        """
        return self._datatype

    @datatype.setter
    def datatype(self, value):
        # This does not differ, as it is a property
        # pylint: disable=arguments-differ
        self._datatype = value

    @property
    def dimensions(self):
        """
        The list of the dimensions in this Output. This will be a tuple of Dimension.
        """
        return self.node.dimensions

    @property
    def fullid(self):
        """
        The full defining ID for the Output
        """
        if self.node is not None:
            return '{}/outputs/{}'.format(self.node.fullid, self.id)
        else:
            return 'fastr://ORPHANED/outputs/{}'.format(self.id)

    @property
    def listeners(self):
        """
        The list of :py:class:`Links <fastr.planning.link.Link>` connected to this Output.
        """
        return self._listeners

    @property
    def preferred_types(self):
        """
        The list of preferred :py:class:`DataTypes <fastr.plugins.managers.datatypemanager.DataType>`
        for this Output.
        """
        if self._preferred_types is not None and len(self._preferred_types) > 0:
            return self._preferred_types
        elif self.node.parent is not None and self.node.parent.preferred_types is not None and len(self.node.parent.preferred_types) > 0:
            return self.node.parent.preferred_types
        else:
            return types.preferred_types

    @preferred_types.setter
    def preferred_types(self, value):
        """
        The list of preferred :py:class:`DataTypes <fastr.plugins.managers.datatypemanager.DataType>`
        for this Output. (setter)
        """
        if isinstance(value, type) and issubclass(value, DataType):
            self._preferred_types = [value]
        elif isinstance(value, list) and all([isinstance(x, type) and issubclass(x, DataType) for x in value]):
            self._preferred_types = value
        else:
            log.warning('Invalid definition of preferred DataTypes, must be a DataType or list of DataTypes! Ignoring!')

    @property
    def valid(self):
        """
        Check if the output is valid, i.e. has a valid cardinality
        """
        return self.check_cardinality()

    @property
    def resulting_datatype(self):
        """
        The :py:class:`DataType <fastr.plugins.managers.datatypemanager.DataType>` that
        will the results of this Output will have.
        """
        requested_types = [l.target.datatype for l in self.listeners if l.target is not None]
        requested_types.append(self.datatype)

        if self.preferred_types is not None and len(self.preferred_types) > 0:
            return types.match_types(requested_types,
                                        preferred=self.preferred_types)
        else:
            return types.match_types(requested_types)

    def _update(self, key, forward=True, backward=False):
        """Update the status and validity of the Output and propagate the update the NodeRun.
        An Output is valid if:

        * the parent NodeRun is valid (see :py:meth:`NodeRun.update <fastr.planning.node.NodeRun.update>`)
        """
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.fullid))

        self.node.update(key, forward, backward)

        if self.node.valid:
            self._status['valid'] = True
        else:
            self._status['valid'] = False
            self._status['messages'] = ['Parent NodeRun is not valid']


class SubOutput(Output):
    """
    The SubOutput is an Output that represents a slice of another Output.
    """

    def __init__(self, output, index):
        """Instantiate a SubOutput

        :param output: the parent output the suboutput slices.
        :param index: the way to slice the parent output
        :type index: int or slice
        :return: created SubOutput
        :raises FastrTypeError: if the output argument is not an instance of :py:class:`Output <fastr.planning.inputoutput.Output>`
        :raises FastrTypeError: if the index argument is not an ``int`` or ``slice``
        """
        if not isinstance(output, Output):
            raise exceptions.FastrTypeError('Second argument for a SubOutput init should be an Output')

        if not isinstance(index, (int, slice)):
            raise exceptions.FastrTypeError('SubOutput index should be an integer or a slice, found ({}, type {})'.format(index, type(index).__name__))

        super(SubOutput, self).__init__(output.node, output.description)
        self.parent = output
        self.index = index

    def __str__(self):
        """
        Get a string version for the SubOutput

        :return: the string version
        :rtype: str
        """
        return '<SubOutput {}>'.format(self.fullid)

    def __getstate__(self):
        """
        Retrieve the state of the SubOutput

        :return: the state of the object
        :rtype dict:
        """
        state = super(SubOutput, self).__getstate__()
        state['index'] = self.indexrep
        return state

    def __setstate__(self, state):
        """
        Set the state of the SubOutput by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        if isinstance(state['index'], str):
            index = [int(x) if len(x) > 0 else None for x in state['index'].split(':')]
            state['index'] = slice(*index)

        state['_preferred_types'] = []
        super(SubOutput, self).__setstate__(state)
        self._preferred_types = None

    def __eq__(self, other):
        """Compare two SubOutput instances with each other. This function ignores
        the parent, node and update status, but tests rest of the dict for equality.
        equality

        :param other: the other instances to compare to
        :type other: SubOutput
        :returns: True if equal, False otherwise
        :rtype: bool
        """
        if not isinstance(other, type(self)):
            return NotImplemented

        dict_self = dict(vars(self))
        del dict_self['_node']
        del dict_self['parent']
        del dict_self['_status']

        dict_other = dict(vars(other))
        del dict_other['_node']
        del dict_other['parent']
        del dict_other['_status']

        return dicteq(dict_self, dict_other)

    def __len__(self):
        """Return the length of the Output.

        .. note::

            In a SubOutput this is always 1.
        """
        return 1

    @property
    def indexrep(self):
        """
        Simple representation of the index.
        """
        if isinstance(self.index, slice):
            index = '{}:{}'.format(self.index.start, self.index.stop)
            index = index.replace('None', '')

            if self.index.step is not None and self.index.step != 1:
                index = '{}:{}'.format(index, self.index.step)
        else:
            index = self.index

        return index

    def cardinality(self):
        """
        Cardinality of this SubOutput depends on the parent Output and ``self.index``

        :param key: key for a specific sample, can be sample index or id
        :type key: tuple of int or :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the cardinality
        :rtype: int, sympy.Symbol, or None
        :raises FastrCardinalityError: if cardinality references an invalid :py:class:`Input <fastr.planning.inputoutput.Input>`
        :raises FastrTypeError: if the referenced cardinality values type cannot be case to int
        :raises FastrValueError: if the referenced cardinality value cannot be case to int
        """
        parent_cardinality = self.parent.cardinality()

        if parent_cardinality is not None:
            if isinstance(parent_cardinality, int):
                if isinstance(self.index, int):
                    if parent_cardinality >= 1:
                        return 1
                    else:
                        return 0
                else:
                    # Calculate the slice effect on a list of length parent cardinality
                    ind_range = self.index.indices(parent_cardinality)
                    return (ind_range[1] - ind_range[0]) // ind_range[2]
            else:
                return parent_cardinality
        else:
            return None

    @property
    def datatype(self):
        """
        The datatype of this SubOutput
        """
        return self.parent.datatype

    @property
    def fullid(self):
        """
        The full defining ID for the SubOutput
        """
        return '{}/{}'.format(self.parent.fullid, self.indexrep)

    @property
    def listeners(self):
        """
        The list of :py:class:`Links <fastr.planning.link.Link>` connected to this Output.
        """
        return self.parent.listeners

    @property
    def node(self):
        """
        The NodeRun to which this SubOutput belongs
        """
        return self.parent.node

    @property
    def preferred_types(self):
        """
        The list of preferred :py:class:`DataTypes <fastr.plugins.managers.datatypemanager.DataType>`
        for this SubOutput.
        """
        return self.parent.preferred_types

    @preferred_types.setter
    def preferred_types(self, value):
        # We need to key for the signature in subclasses, shut pylint up
        # pylint: disable=unused-argument,no-self-use,arguments-differ
        raise exceptions.FastrNotImplementedError('Cannot set DataType of SubOutput, use the parent Output instead')

    @property
    def samples(self):
        """
        The :py:class:`SampleCollection <fastr.core.sampleidlist.SampleCollection>`
        for this SubOutput
        """
        return self.parent.samples

    @property
    def resulting_datatype(self):
        """
        The :py:class:`DataType <fastr.plugins.managers.datatypemanager.DataType>` that
        will the results of this SubOutput will have.
        """
        return self.parent.resulting_datatype

    def _update(self, key, forward=True, backward=False):
        """Update the status and validity of the SubOutput and propagate the update downstream.
        An SubOutput is valid if:

        * the parent NodeRun is valid (see :py:meth:`NodeRun.update <fastr.planning.node.NodeRun.update>`)

        """
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.fullid))
        self.parent.update(key, forward, backward)

        if self.node.valid:
            self._status['valid'] = True
        else:
            self._status['valid'] = False
            self._status['messages'] = ['Parent NodeRun is not valid']


class AdvancedFlowOutput(Output):
    """
    Output for nodes that have an advanced flow. This means that the output
    sample id and index is not the same as the input sample id and index.
    The AdvancedFlowOutput has one extra dimensions that is created by the
    Node.
    """
    @property
    def dimensions(self):
        parent_dimensions = super(AdvancedFlowOutput, self).dimensions

        return tuple(Dimension('{}_{}'.format(d.name, self.id), d.size) for d in parent_dimensions[:-1]) + (parent_dimensions[-1],)


class MacroOutput(Output):
    @property
    def dimensions(self):
        return self.node.get_output_info(self)


class SourceOutput(Output):
    """
    Output for a SourceNodeRun, this type of Output determines the cardinality in
    a different way than a normal NodeRun.
    """
    def __init__(self, node, description):
        """Instantiate a FlowOutput

        :param node: the parent node the output belongs to.
        :param description: the :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
               describing the output.
        :return: created FlowOutput
        :raises FastrTypeError: if description is not of class
                                :py:class:`ParameterDescription <fastr.core.tool.ParameterDescription>`
        :raises FastrDataTypeNotAvailableError: if the DataType requested cannot be found in the ``types``
        """
        super(SourceOutput, self).__init__(node, description)

        self._linearized = None

    @property
    def linearized(self):
        """
        A linearized version of the sample data, this is lazily cached
        linearized version of the underlying SampleCollection.
        """
        if self._linearized is None:
            self._linearized = tuple(self.samples[x] for x in self.samples)

        return self._linearized

    def cardinality(self):
        """
        Cardinality of this SourceOutput, may depend on the inputs of the parent NodeRun.

        :param key: key for a specific sample, can be sample index or id
        :type key: tuple of int or :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the cardinality
        :rtype: int, sympy.Symbol, or None
        """
        return sympy.symbols('N_{}'.format(self.node.id.replace(' ', '_')))



import fastr
import failing_network


def create_network():
    # Create network to embed
    child_network = failing_network.create_network()

    # Create Network
    network = fastr.create_network('failing_macro_top_level')

    # create sources
    source_a = network.create_source('Int', id='source_a')
    source_b = network.create_source('Int', id='source_b')
    source_c = network.create_source('Int', id='source_c')

    # Create MacroNode
    failing_macro = network.create_macro(child_network, id='failing_macro')

    # Link sources to macro
    failing_macro.inputs['source_1'] = source_a.output
    failing_macro.inputs['source_2'] = source_b.output
    failing_macro.inputs['source_3'] = source_c.output

    # Addint
    add = network.create_node('fastr/math/AddInt:1.0', tool_version='1.0', id="add")

    # Create sink
    sink = network.create_sink('Int', id='sink')

    # Adjust constants(non required inputs) in macro network
    #  add_multiple_ints_node.inputs['const_addint1_value2__add_multiple_ints_1'] = 1234, #input.output
    # Link the network
    add.inputs['left_hand'] = failing_macro.outputs['sink_5']
    add.inputs['right_hand'] = (10,)

    sink.inputs['input'] = add.outputs['result']

    return network


def source_data(network):
    fastr.log.info('Creating source data for {}'.format(network.id))
    return {
        'source_a': {'sample_1': 'vfslist://example_data/add_ints/values'},
        'source_b': {'sample_1': 'vfslist://example_data/add_ints/values'},
        'source_c': {'sample_1': 'vfslist://example_data/add_ints/values'},
    }


def sink_data(network):
    return {'sink': 'vfs://tmp/results/{}/result_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id)}


def main():
    network = create_network()

    # Execute
    network.draw(network.id, draw_dimensions=True, expand_macros=True)
    network.execute(source_data(network), sink_data(network))


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABCMeta, abstractmethod
from collections import deque
import datetime
import getpass
import itertools
import json
from logging import LogRecord
import os
import time
from typing import Dict
from threading import Lock, Thread

import requests
import fastr

from fastr.helpers.classproperty import classproperty
from fastr.plugins.reportingplugin import ReportingPlugin
from fastr.execution.job import JobState, Job
from fastr.execution.networkrun import NetworkRun


class BasePimAPI(metaclass=ABCMeta):
    """
    Base class for PIM API classes which specifies the methods required to function
    """
    @abstractmethod
    def pim_update_status(self, job: Job):
        """
        Update the status of a job

        :param job: The job which to update
        """

    @abstractmethod
    def pim_register_run(self, network: NetworkRun):
        """
        Send the basic Network layout to PIM and register the run.

        :param network: The network run to register to PIM
        """

    @abstractmethod
    def pim_finish_run(self, network: NetworkRun):
        """
        Set the PIM run to finished and clean up

        :param network: The network run to finish
        """

    @abstractmethod
    def pim_log_line(self, record: LogRecord):
        """
        Send a new line of log record to PIM
        :param record: the log record to send
        """


class PimAPIv2(object):
    """
    Class to publish to PIM
    """
    PIM_STATUS_MAPPING = {
        JobState.nonexistent: 5,
        JobState.created: 0,
        JobState.queued: 0,
        JobState.hold: 0,
        JobState.running: 1,
        JobState.execution_done: 1,
        JobState.execution_failed: 1,
        JobState.processing_callback: 1,
        JobState.finished: 2,
        JobState.failed: 3,
        JobState.cancelled: 4,
    }

    NODE_CLASSES = {
        'NodeRun': 'node',
        'SourceNodeRun': 'source',
        'ConstantNodeRun': 'constant',
        'SinkNodeRun': 'sink'
    }

    STATUS_TYPES = [
        {
            "color": "#aaccff",
            "description": "Jobs that are waiting for input",
            "title": "idle"
        },
        {
            "color": "#daa520",
            "description": "Jobs that are running",
            "title": "running"
        },
        {
            "color": "#23b22f",
            "description": "Jobs that finished successfully",
            "title": "success"
        },
        {
            "color": "#dd3311",
            "description": "Jobs that have failed",
            "title": "failed"
        },
        {
            "color": "#334477",
            "description": "Jobs which were cancelled",
            "title": "cancelled"
        },
        {
            "color": "#ccaa99",
            "description": "Jobs with an undefined state",
            "title": "undefined"
        }
    ]

    def __init__(self, uri=None):
        self.pim_uri = uri
        self.registered = False
        self.run_id = None
        self.jobs_uri = None

        # Data for sending will be cached and flushed
        self.running = True
        self.submit_thread = Thread(target=self.job_update_loop, name='PimSubmitter', daemon=True)
        self.update_interval = fastr.config.pim_update_interval
        self.batch_size = fastr.config.pim_batch_size
        self.finished_timeout = fastr.config.pim_finished_timeout

        self.queued_job_updates = deque()
        self.submitted_job_updates = []
        self.updates_lock = Lock()

        # Some data
        self.counter = itertools.count()
        self.scopes = {None: 'root'}
        self.nodes = {}
        self.job_states = {}

    def create_job_data(self, job: Job) -> Dict:
        """
        Create a job data json part that is ready to send to PIM

        :param job: the job to convert
        :return:
        """
        try:
            node = self.nodes[job.node_global_id]
        except KeyError:
            fastr.log.info('NODES: {}'.format(self.nodes))
            raise

        # Create PIM job data
        pim_job_data = {
            "path": '{}/{}'.format(node, job.id),
            "title": "",
            "customData": {
                "sample_id": list(job.sample_id),
                "sample_index": list(job.sample_index),
                "errors": job.errors,
            },
            "status": self.PIM_STATUS_MAPPING[job.status],
            "description": "",
        }

        # Send the data to PIM
        fastr.log.debug('Updating PIM job status {} => {} ({})'.format(job.id,
                                                                       job.status,
                                                                       self.PIM_STATUS_MAPPING[job.status]))

        if (job.status == JobState.failed or fastr.config.pim_debug) and os.path.exists(job.extrainfofile):

            with open(job.extrainfofile) as extra_info_file:
                extra_info = json.load(extra_info_file)

                process = extra_info.get('process')

                if process:
                    # Process information
                    pim_job_data['customData']['__stdout__'] = process.get('stdout')
                    pim_job_data['customData']['__stderr__'] = process.get('stderr')
                    pim_job_data['customData']['command'] = process.get('command')
                    pim_job_data['customData']['returncode'] = process.get('returncode')
                    pim_job_data['customData']['time_elapsed'] = process.get('time_elapsed')

                # Host information
                pim_job_data['customData']['hostinfo'] = extra_info.get('hostinfo')

                # Input output hashes for validation of files
                pim_job_data['customData']['input_hash'] = extra_info.get('input_hash')
                pim_job_data['customData']['output_hash'] = extra_info.get('output_hash')

                # Tool information
                pim_job_data['customData']['tool_name'] = extra_info.get('tool_name')
                pim_job_data['customData']['tool_version'] = extra_info.get('tool_version')

        return pim_job_data

    def pim_update_status(self, job):
        if self.pim_uri is None:
            return

        if not self.registered:
            fastr.log.debug('Did not register a RUN with PIM yet! Cannot'
                            ' send status updates!')
            return

        if job.status not in [
            JobState.created,
            JobState.running,
            JobState.finished,
            JobState.cancelled,
            JobState.failed,
        ]:
            return

        if job.id in self.job_states and self.job_states[job.id] == self.PIM_STATUS_MAPPING[job.status]:
            # Not a valid update
            fastr.log.debug('Ignoring non-PIM update')
            return
        else:
            self.job_states[job.id] = self.PIM_STATUS_MAPPING[job.status]

        with self.updates_lock:
            self.queued_job_updates.append(self.create_job_data(job))

    def job_update_loop(self):
        """
        Loop that periodically updates job data
        """
        while self.running or len(self.submitted_job_updates) != 0 or len(self.queued_job_updates) != 0:
            # Not time of start of this loop iteration
            last_update = time.time()

            # Select job data to submit
            with self.updates_lock:
                # Find number of jobs we can add
                number_of_jobs = min(self.batch_size - len(self.submitted_job_updates),
                                     len(self.queued_job_updates))

                # Append extra jobs to fill the batch size
                for _ in range(number_of_jobs):
                    self.submitted_job_updates.append(
                        self.queued_job_updates.popleft()
                    )

            # If there are jobs update to submit, do so
            if len(self.submitted_job_updates) > 0:
                try:
                    response = requests.put(self.jobs_uri, json=self.submitted_job_updates, timeout=5)
                except requests.ConnectionError as exception:
                    fastr.log.warning('Could no publish status to PIM, encountered exception: {}'.format(exception))
                except requests.Timeout:
                    fastr.log.warning('Connection to PIM timed out during job update submission')
                else:
                    if response.status_code >= 300:
                        fastr.log.warning('Response of jobs update: [{r.status_code}] {r.text}'.format(r=response))
                    else:
                        # Success
                        self.submitted_job_updates = []

            # Sleep to get a roughly update_interval second loop time
            time_to_sleep = last_update + self.update_interval - time.time()
            time_to_sleep = time_to_sleep if time_to_sleep > 0 else 0
            time.sleep(time_to_sleep)

    def pim_serialize_node(self, node, scope, links):
        # Fish out macros and use specialized function
        if type(node).__name__ == 'MacroNodeRun':
            return self.pim_serialize_macro(node, scope, links)

        node_data = {
            "name": node.id,
            "title": node.id,
            "children": [],
            "customData": {},
            "inPorts": [
                {
                    "name": output.id,
                    "title": output.id,
                    "customData": {
                        "input_group": output.input_group,
                        "datatype": output.datatype.id,
                        "dimension_names": [x.name for x in output.dimensions],
                    },
                } for output in node.inputs.values()
            ],
            "outPorts": [
                {
                    "name": output.id,
                    "title": output.id,
                    "customData": {
                        "datatype": output.resulting_datatype.id,
                        "dimension_names": [x.name for x in output.dimensions],
                    },
                } for output in node.outputs.values()
            ],
            "type": type(node).__name__,
        }

        if type(node).__name__ == 'SourceNodeRun':
            node_data['inPorts'].append(
                {
                    "name": 'input',
                    "title": 'input',
                    "customData": {
                        "datatype": node.output.datatype.id,
                        "dimension_names": [node.id],
                    },
                }
            )

        if type(node).__name__ == 'SinkNodeRun':
            node_data['outPorts'].append(
                {
                    "name": 'output',
                    "title": 'output',
                    "customData": {
                        "datatype": node.input.datatype.id,
                        "dimension_names": node.dimnames,
                    },
                }
            )

        # Register node id mapping
        self.nodes[node.global_id] = '{}/{}'.format(scope, node.id)

        return node_data

    def pim_serialize_macro(self, node, scope, links):
        new_scope = '{}/{}'.format(scope, node.id)
        self.nodes[node.global_id] = new_scope

        # Set node data
        node_data = {
            "name": node.id,
            "title": node.id,
            "children": [],
            "customData": {},
            "inPorts": [],
            "outPorts": [],
            "type": type(node).__name__,
        }

        # Serialize underlying network
        self.pim_serialize_network(node.network_run, new_scope, node_data, links)

        return node_data

    def pim_serialize_network(self, network, scope, parent, links):
        visited_nodes = set()

        for step_name, step_nodes in network.stepids.items():
            step_data = {
                "name": step_name,
                "title": step_name,
                "children": [],
                "customData": {},
                "inPorts": [],
                "outPorts": [],
                "type": "NetworkStep",
            }

            parent['children'].append(step_data)

            # Serialize nodes to parents child list
            for node in step_nodes:
                step_data['children'].append(self.pim_serialize_node(node, '{}/{}'.format(scope, step_name), links))
                visited_nodes.add(node.id)

        # Serialize nodes to parents child list
        for node in network.nodelist.values():
            if node.id not in visited_nodes:
                parent['children'].append(self.pim_serialize_node(node, scope, links))

        # Serialize links to global link list
        for link in network.linklist.values():
            links.append(self.pim_serialize_link(link))

    def pim_serialize_link(self, link):
        if type(link.source.node).__name__ == 'MacroNodeRun':
            from_port = "{}/{}/output".format(self.nodes[link.source.node.global_id],
                                              link.source.id)
        else:
            from_port = "{}/{}".format(self.nodes[link.source.node.global_id],
                                       link.source.id)

        if type(link.target.node).__name__ == 'MacroNodeRun':
            to_port = "{}/{}/input".format(self.nodes[link.target.node.global_id],
                                           link.target.id)
        else:
            to_port = "{}/{}".format(self.nodes[link.target.node.global_id],
                                     link.target.id)

        link_data = {
            "customData": {
                "expand": link.expand,
                "collapse": link.collapse
            },
            "description": "",
            "fromPort": from_port,
            "name": link.id,
            "title": link.id,
            "toPort": to_port,
            "dataType": link.source.resulting_datatype.id
        }

        return link_data

    def pim_register_run(self, network):
        if self.pim_uri is None:
            fastr.log.warning('No valid PIM uri known. Cannot register to PIM!')
            return

        self.run_id = network.id
        pim_run_data = {
            "title": self.run_id,
            "name": self.run_id,
            "assignedTo": [],
            "user": fastr.config.pim_username,
            "root": {
                "name": "root",
                "title": network.network_id,
                "description": "",
                "children": [],
                "customData": {},
                "inPorts": [],
                "outPorts": [],
                "type": "NetworkRun",
            },
            "links": [],
            "description": "Run of {} started at {}".format(network.id,
                                                            network.timestamp),
            "customData": {
                "workflow_engine": "fastr",
                "tmpdir": network.tmpdir,
            },
            "statusTypes": self.STATUS_TYPES,
        }

        self.pim_serialize_network(network=network,
                                   scope="root",
                                   parent=pim_run_data["root"],
                                   links=pim_run_data["links"])

        uri = '{pim}/api/runs/'.format(pim=fastr.config.pim_host)
        fastr.log.info('Registering {} with PIM at {}'.format(self.run_id, uri))

        fastr.log.debug('Send PUT to pim at {}:\n{}'.format(uri, json.dumps(pim_run_data, indent=2)))

        # Cache the jobs uri
        self.jobs_uri = '{pim}/api/runs/{run_id}/jobs'.format(pim=fastr.config.pim_host,
                                                              run_id=self.run_id)

        # Send out the response and record if we registered correctly
        try:
            response = requests.post(uri, json=pim_run_data)
            if response.status_code in [200, 201]:
                self.registered = True
                self.submit_thread.start()
                fastr.log.info('Run registered in PIM at {}/runs/{}'.format(fastr.config.pim_host,
                                                                            self.run_id))
            else:
                fastr.log.warning('Could not register run at PIM, got a {} response'.format(response.status_code))
                fastr.log.warning('Response: {}'.format(response.text))
        except requests.ConnectionError as exception:
            fastr.log.error('Could no register network to PIM, encountered'
                            ' exception: {}'.format(exception))

        # Create dummy job
        root_job_data = {
            "path": 'root/master',
            "title": "",
            "customData": {
            },
            "status": self.PIM_STATUS_MAPPING[JobState.running],
            "description": "",
        }

        self.queued_job_updates.append(root_job_data)

    def pim_finish_run(self, run):
        # Finish dummy job
        root_job_data = {
            "path": 'root/master',
            "title": "",
            "customData": {
            },
            "status": self.PIM_STATUS_MAPPING[JobState.finished if run.result else JobState.failed],
            "description": "",
        }
        self.queued_job_updates.append(root_job_data)

        self.running = False

        run_finished = datetime.datetime.now()

        while len(self.queued_job_updates) != 0 or len(self.submitted_job_updates) != 0:
            timer = datetime.datetime.now() - run_finished
            if timer.seconds > self.finished_timeout:
                fastr.log.warning('Not all PIM updates sent, timeout reached!')
                break
            time.sleep(self.update_interval)
            fastr.log.info('Waiting for all jobs to be published to PIM...')

    def pim_log_line(self, record: LogRecord):
        timestamp = datetime.datetime.utcfromtimestamp(record.created).strftime('%Y-%m-%d %H:%M:%S.%f')

        root_job_data = {
            "path": 'root/master',
            "customData": {
                '__log__': {
                    timestamp: {
                        'process_name': record.processName,
                        'thread_name': record.threadName,
                        'level_name': record.levelname,
                        'module': record.module,
                        'function': record.funcName,
                        'lineno': record.lineno,
                        'message': record.msg
                    }
                }
            }
        }
        self.queued_job_updates.append(root_job_data)


class PimReporter(ReportingPlugin):
    SUPPORTED_APIS = {
        2: PimAPIv2,
    }

    def __init__(self):
        self.pim_uri = None
        super().__init__()
        self.api = None

    def activate(self):
        """
        Activate the reporting plugin
        """
        super().activate()

        # Parse URI
        if fastr.config.pim_host == '':
            fastr.log.info("No valid PIM host given, PIM publishing will be disabled!")
            self.pim_uri = None
            self.api = None
            return

        self.pim_uri = fastr.config.pim_host

        try:
            response = requests.get('{pim}/api/info'.format(pim=self.pim_uri))
            if response.status_code >= 300:
                version = 1
            else:
                version = response.json().get('version', 1)
        except requests.ConnectionError as exception:
            fastr.log.error('Could no publish status to PIM, encountered exception: {}'.format(exception))
            return

        try:
            api_class = self.SUPPORTED_APIS[version]
            fastr.log.info('Using PIM API version {}'.format(version))
        except KeyError:
            fastr.log.error('PIM API version {} not supported!'.format(version))
            return

        self.api = api_class(self.pim_uri)

    @classproperty
    def configuration_fields(cls):
        return {
            "pim_host": (str, "", "The PIM host to report to"),
            "pim_username": (str, getpass.getuser(), "Username to send to PIM",
                             "Username of the currently logged in user"),
            "pim_update_interval": (float, 2.5, "The interval in which to send jobs to PIM"),
            "pim_batch_size": (int, 100, "Maximum number of jobs that can be send to PIM in a single interval"),
            "pim_debug": (bool, False, "Setup PIM debug mode to send stdout stderr on job success"),
            "pim_finished_timeout": (int, 10, "Maximum number of seconds after the network finished in which PIM "
                                              "tries to synchronize all remaining jobs")
        }

    def job_updated(self, job: Job):
        if self.pim_uri and self.api:
            self.api.pim_update_status(job)

    def run_started(self, run: NetworkRun):
        if self.pim_uri and self.api:
            self.api.pim_register_run(run)

    def run_finished(self, run: NetworkRun):
        if self.pim_uri and self.api:
            self.api.pim_finish_run(run)

    def log_record_emitted(self, record: LogRecord):
        if self.pim_uri and self.api and self.api.running:
            self.api.pim_log_line(record)




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains both the Updateable class and UpdateableMeta meta-class
for objects which support updates within the fastr system
"""

from abc import ABCMeta, abstractmethod
import os
import types
import threading

from .. import exceptions
from ..helpers import log


class UpdateableMeta(ABCMeta):
    """
    A metaclass for objects which are updateable and need some methods/properties
    to trigger an update.
    """

    @classmethod
    def calcmro(mcs, bases):
        """Calculate the Method Resolution Order of bases using the C3 algorithm.

        Suppose you intended creating a class K with the given base classes. This
        function returns the MRO which K would have, *excluding* K itself (since
        it doesn't yet exist), as if you had actually created the class.

        Another way of looking at this, if you pass a single class K, this will
        return the linearization of K (the MRO of K, *including* itself).

        :param bases: the list of bases for which create the MRO
        :return: the list representing the entire MRO, except the (non-existing) class itself

        Note: Taken from http://code.activestate.com/recipes/577748-calculate-the-mro-of-a-class/
              Created by Steven D'Aprano and licensed under the MIT license
        """
        seqs = [list(C.__mro__) for C in bases] + [list(bases)]
        res = []
        while True:
            non_empty = [item for item in seqs if item]
            if not non_empty:
                # Nothing left to process, we're done.
                return tuple(res)
            for seq in non_empty:  # Find merge candidates among seq heads.
                candidate = seq[0]
                not_head = [s for s in non_empty if candidate in s[1:]]
                if not_head:
                    # Reject the candidate.
                    candidate = None
                else:
                    break
            if not candidate:
                raise TypeError("inconsistent hierarchy, no C3 MRO is possible")
            res.append(candidate)
            for seq in non_empty:
                # Remove candidate.
                if seq[0] == candidate:
                    del seq[0]

    @classmethod
    def find_member(mcs, name, parents, dct):
        """
        Find  a member of the class in the same way as Python would if it had a given dict and set of bases

        :param mcs: metaclass at work
        :param name: name of the class to be created
        :param parents: list of the bases for the new class
        :param dct: the dict of the class being created
        :return: the firstly resolved member or None if nothing found
        """
        if name in dct:
            return dct[name]

        for cls in mcs.calcmro(parents):
            if hasattr(cls, name):
                return getattr(cls, name)

        return None

    def __new__(mcs, name, parents, dct):
        if '__updatetriggers__' in dct:
            triggers = dct['__updatetriggers__']

            for trigger in triggers:
                fnc = mcs.find_member(trigger, parents, dct)

                if fnc is not None:
                    if isinstance(fnc, (types.FunctionType, types.MethodType, types.BuiltinMethodType)):
                        log.debug('Adding update trigger {} to {}'.format(trigger, name))
                        dct[trigger] = mcs.updatetrigger(fnc)
                    else:
                        log.debug('Skipping trigger {} for {} (wrong type: {})'.format(trigger, name, type(fnc).__name__))
                else:
                    log.debug('Skipping trigger {} for {} (not in dct)'.format(trigger, name))
                    raise ValueError()

        return super(UpdateableMeta, mcs).__new__(mcs, name, parents, dct)

    @staticmethod
    def updatetrigger(fnc):
        """
        Function decorator to make a function trigger an update after being
        called. This is a way to easily have function trigger an update
        after setting a value without writing tons of wrapper functions.
        The function keeps the original docstring and appends a note to it.
        """
        def wrapper(self, *args, **kwargs):
            """
            Decorator wrapper around a function. This docstring will be
            changed for each wrapped function
            """
            # First call the wrapped function
            fnc(self, *args, **kwargs)
            # Call the update function
            # This function name is chosen so that it has low chance of conflicts
            if self.__updating__:
                self.__updatefunc__()

        # Add a note to the docstring indicating what is going on
        original = '{}.{}'.format(fnc.__module__, fnc.__name__)
        docstring = fnc.__doc__.lstrip('\n')
        indent = docstring[:len(docstring) - len(docstring.lstrip())]
        extra_doc = ('\n.. note::\n'
                     '    This is a wrapped version of ``{orig}``\n'
                     '    which triggers an update of  the object after\n'
                     '    being called').format(orig=original)
        extra_doc = '\n{}'.format(indent).join(extra_doc.splitlines())
        wrapper.__doc__ = docstring.rstrip() + "\n{indent}\n{indent}{extra_doc}\n{indent}".format(indent=indent, extra_doc=extra_doc)
        return wrapper


class Updateable(object, metaclass=UpdateableMeta):

    """
    Super class for all classes that can be updated and have a status.
    These objects can be valid/invalid state. These states are set by
    the function update. This allows for interactively checking the network.
    """

    #: Which methods need to be wrapped to trigger an update. Override this
    #: value to have the functions automatically wrapped. E.g.
    #: ``__update_triggers__ = ['append', 'insert', '__setitem__']`` to have
    #: these functions wrapped.
    __updatetriggers__ = []

    #: Flag to indicate that this object is allowed to update
    __updating__ = True

    #: Lock to avoid multiple updates happening at the same time
    __updateinprogress__ = threading.Lock()

    def __init__(self):
        """
        Constructor, creates the status field

        :return: newly created object
        """
        self._status = {'key': None, 'valid': None, 'messages': []}

    def __getstate__(self):
        """
        Retrieve the state of the object, make sure the status is not part of
        the description as it will not be valid after re-creating the object.

        :return: the state of the object
        :rtype dict:
        """
        return {}

    def __setstate__(self, state):
        """
        Set the state of the object by the given state. This adds a clean
        status field, making sure it is not unintended, outdated information
        from before serialization.

        :param dict state: The state to populate the object with
        """
        self._status = {'key': None, 'valid': None, 'messages': []}

    @property
    def messages(self):
        """
        The messages of the last update
        """
        return self._status['messages']

    @property
    def valid(self):
        """
        Flag indicating that the object is valid
        """
        return self._status['valid']

    def update(self, key=None, forward=True, backward=False):
        """
        Default function for updating, it can be called without key to have a
        new update started with a new key.

        :param int key: a key for this update, should be different than the
                        last update key
        :param bool forward: flag indicating to update forward in the network
        :param bool backward: flag indicating to update backward in the network
        """
        # If updating is disable, we don't do anything
        if not self.__updating__:
            return

        # If this is the first update function called in this round, create a key
        lock = False
        if key is None:
            key = hash(os.urandom(128))

            # Block update until previous updates are finished
            log.debug('Getting update lock')
            lock = Updateable.__updateinprogress__.acquire()
            log.debug('Got update lock')

        # Only update if not update in this update round
        if key == self._status['key']:
            # Release updatable lock, this should never be needed, but if not done when needed causes a deadlock
            if lock:
                log.warning('Releasing lock in unexpected place! Might be a bug!')
                Updateable.__updateinprogress__.release()
                log.debug('Released update lock')
            return

        # Store key to mark last update round
        self._status['key'] = key

        # Run the update func
        self._update(key, forward, backward)

        # Release updatable lock
        if lock:
            Updateable.__updateinprogress__.release()
            log.debug('Released update lock')

        return key

    # Set the update function for the trigger functions to call
    __updatefunc__ = update

    @abstractmethod
    def _update(self, key, forward=True, backward=False):
        """
        The actual update function to be used. This is an abstract function
        that Updateable objects must implement.

        :param int key: a key for this update, should be different than the
                        last update key
        :param bool forward: flag indicating to update forward in the network
        :param bool backward: flag indicating to update backward in the network
        """
        raise exceptions.FastrNotImplementedError('Purposefully not implemented')



import fastr


def create_network():
    network = fastr.create_network('auto_prefix_test')

    const = network.create_constant(fastr.datatypes.Int, [1, 2, 3], id='const')
    source = network.create_source(fastr.datatypes.Int, id='source')

    node_m_p = network.create_node('fastr/util/AutoPrefixTest:1.0', tool_version='1.0', id='m_p')
    node_a_p = network.create_node('fastr/util/AutoPrefixTest:1.0', tool_version='1.0', id='a_p')
    node_m_n = network.create_node('fastr/util/AutoPrefixNegateTest:1.0', tool_version='1.0', id='m_n')
    node_a_n = network.create_node('fastr/util/AutoPrefixNegateTest:1.0', tool_version='1.0', id='a_n')

    sink_m_p = network.create_sink('Int', id='sink_m_p')
    sink_a_p = network.create_sink('Int', id='sink_a_p')
    sink_m_n = network.create_sink('Int', id='sink_m_n')
    sink_a_n = network.create_sink('Int', id='sink_a_n')

    for node in [node_m_p, node_a_p, node_m_n, node_a_n]:
        node.inputs['left_hand'] = source.output
        node.inputs['right_hand'] = const.output

    sink_m_p.input = node_m_p.outputs['multiplied']
    sink_m_n.input = node_m_n.outputs['multiplied']
    sink_a_n.input = node_a_n.outputs['added']
    sink_a_p.input = node_a_p.outputs['added']

    return network


def source_data(network):
    return {'source': [5]}


def sink_data(network):
    return {
        'sink_a_n': 'vfs://tmp/results/{network}/{{node}}_{{sample_id}}'.format(network=network.id),
        'sink_a_p': 'vfs://tmp/results/{network}/{{node}}_{{sample_id}}'.format(network=network.id),
        'sink_m_p': 'vfs://tmp/results/{network}/{{node}}_{{sample_id}}'.format(network=network.id),
        'sink_m_n': 'vfs://tmp/results/{network}/{{node}}_{{sample_id}}'.format(network=network.id),
    }


def main():
    network = create_network()
    network.save('auto_prefix_test.json')

    # Execute
    network.draw(network.id, draw_dimensions=True)
    network.execute(source_data(network), sink_data(network))


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle
import os

import pytest

import fastr
from fastr import datatypes
from fastr import exceptions

DATA_DIR = os.path.join(os.path.dirname(__file__),
                        'data')


def test_equal():
    """
    Test if the equal operator functions correct
    """
    # Test if equal operator works well
    assert datatypes.Int(42) == datatypes.Int(42)
    assert datatypes.Float(2.5) == datatypes.Float(2.5)
    assert datatypes.String('TestString') == datatypes.String('TestString')

    # Test a few wrong cases
    assert not (datatypes.Int(42) == datatypes.Int(24))
    assert not (datatypes.String('TestString') == datatypes.String('teststring'))
    assert not (datatypes.Int(42) == datatypes.Float(42))


def test_unequal():
    """
    Test if the unequal operator functions correct
    """
    # Test if equal operator works well
    assert not (datatypes.Int(42) != datatypes.Int(42))
    assert not (datatypes.Float(2.5) != datatypes.Float(2.5))
    assert not (datatypes.String('TestString') != datatypes.String('TestString'))

    # Test a few wrong cases
    assert datatypes.Int(42) != datatypes.Int(24)
    assert datatypes.String('TestString') != datatypes.String('teststring')
    assert datatypes.Int(42) != datatypes.Float(42)


def test_load_pickle():
    """
    Load some data from pickles and test it against predefined values
    """
    with open(os.path.join(DATA_DIR, 'int_42.pickle'), 'rb') as fin:
        test_int = pickle.load(fin)
    assert test_int == datatypes.Int(42)

    with open(os.path.join(DATA_DIR, 'float_pi.pickle'), 'rb') as fin:
        test_float = pickle.load(fin)
    assert test_float == datatypes.Float(3.14159265359)

    with open(os.path.join(DATA_DIR, 'enum_test.pickle'), 'rb') as fin:
        test_enum = pickle.load(fin)

    enum_type = fastr.types.create_enumtype('TestEnumSerialization',
                                            options=['fails', 'writes', 'loads'])
    assert test_enum == enum_type('loads')


def test_getstate_setstate():
    """
    Test if the getstate and set state work well
    """
    # Test some known states
    assert datatypes.Int(3).__getstate__() == ('Int', 3, None)
    assert datatypes.Float(4.2).__getstate__() == ('Float', 4.2, None)
    assert datatypes.String('test').__getstate__() == ('String', 'test', None)

    # Do a round trip
    source_int = datatypes.Int(10)
    target_int = datatypes.Int()

    target_int.__setstate__(source_int.__getstate__())
    assert source_int == target_int


def test_wrong_setstate():
    """
    Test if the TypeCheck in get/set state works correctly
    """
    test_float = datatypes.Float(42)
    test_int = datatypes.Int()
    with pytest.raises(exceptions.FastrValueError):
        test_int.__setstate__(test_float.__getstate__())


def test_round_trip_pickle():
    """
    Dump datatype objects to a pickle and rebuilt them. Verify that they are the same.
    """
    # Round trip with an Int
    test_int = datatypes.Int(7)
    test_int_rebuilt = pickle.loads(pickle.dumps(test_int))
    assert test_int == test_int_rebuilt

    # Round trip with an Enum
    enum_type = fastr.types.create_enumtype('TestEnumSerializationRoundTrip',
                                            options=['ham', 'spam', 'eggs'])
    test_enum = enum_type('eggs')
    test_enum_rebuilt = pickle.loads(pickle.dumps(test_enum))
    assert test_enum == test_enum_rebuilt


def test_raw_value():
    test_string = datatypes.String('Test string')
    assert test_string.value == test_string.raw_value


def test_validate_error():
    class WrongDataType(datatypes.URLType):
        def _validate(self):
            raise NotImplementedError("We don't like no validation!")

    test_data = WrongDataType('test')
    with pytest.raises(NotImplementedError):
        print('{} is {}'.format(test_data,
                                'valid' if test_data.valid else 'not valid'))


def test_default_validate():
    class SimpleDatatype(datatypes.BaseDataType):
        def __init__(self, value):
            super(SimpleDatatype, self).__init__(value)

    test_data = SimpleDatatype('test 123')
    print('{} is {}'.format(test_data,
                            'valid' if test_data.valid else 'not valid'))
    assert test_data.valid


def test_checksum():
    """
    Make sure some type checksums stay constant, including some duplicated
    files.
    """
    img1_path = os.path.join(fastr.config.mounts['example_data'], 'images', 'mrwhite.mhd')
    img2_path = os.path.join(fastr.config.mounts['example_data'], 'images', 'mrwhite_duplicate.mhd')
    txt1_path = os.path.join(fastr.config.mounts['example_data'], 'add_ints', 'values')
    txt2_path = os.path.join(fastr.config.mounts['example_data'], 'add_ints', 'values_duplicate')

    assert datatypes.Int(42).checksum() == 'a1d0c6e83f027327d8461063f4ac58a6'
    assert datatypes.Float('13.37').checksum() == '4cf47d0124912e13713bc06b31971c5e'
    assert datatypes.String('Test123').checksum() == '68eacb97d86f0c4621fa2b0e17cabd8c'
    assert datatypes.TxtFile(txt1_path).checksum() == 'b742e35f18a05cf2ebac5e3f0a91599b'
    assert datatypes.TxtFile(txt2_path).checksum() == 'b742e35f18a05cf2ebac5e3f0a91599b'
    assert datatypes.MetaImageFile(img1_path).checksum() == 'eabeb2aca52d581f701070eb84c1934b'
    assert datatypes.MetaImageFile(img2_path).checksum() == 'eabeb2aca52d581f701070eb84c1934b'



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib


def clean():
    """
    Clean the _autogen and main directories of .rst and .txt files
    """
    doc_path = pathlib.Path(__file__).parent
    clean_dir(doc_path)

    autogen_path = doc_path / '_autogen'
    clean_dir(autogen_path)


def clean_dir(directory):
    for filepath in directory.iterdir():
        if filepath.suffix in ('.rst', '.txt'):
            filepath.unlink()
            print(f'removed {filepath}')


if __name__ == '__main__':
    clean()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta

from .. import exceptions
from ..abc.serializable import Serializable
from ..abc.updateable import Updateable


class BaseNodeRun(Updateable, Serializable, metaclass=ABCMeta):
    NODE_RUN_TYPES = {}
    NODE_RUN_MAP = {}

    def __init_subclass__(cls, **kwargs):
        """
        Register nodes in class for easly location
        """
        name = cls.__name__
        cls.NODE_RUN_TYPES[name] = cls
        if name.endswith('Run'):
            key = name[:-3]
        else:
            raise exceptions.FastrValueError('Cannot determine matching Node class for this BaseNodeRun subclass.')

        cls.NODE_RUN_MAP[key] = cls



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from collections import namedtuple
from copy import deepcopy

import fastr
from fastr import exceptions
from fastr.core.version import Version
from fastr.utils.cmd import add_parser_doc_link
FASTR_LOG_TYPE = 'none'  # Do not get info about fastr


class dummy_container:
    def __getitem__(self, value):
        return value


FastrNamespaceType = namedtuple('FastrNamespaceType', ['toollist', 'typelist'])


EVAL_GLOBALS = {
    'fastr': FastrNamespaceType(dummy_container(), dummy_container()),
    'toollist': dummy_container(),
    'typelist': dummy_container(),
}


def get_parser():
    parser = argparse.ArgumentParser(
        description=("Upgrades a python file that creates a Network to the new"
                     " fastr 3.x syntax. The file will be parsed and the"
                     " full syntax tree will be transformed to fit the new"
                     " syntax."
                     "\n\n.. note:: Solves most common problems, but cannot always solve 100% of the issues")
    )
    parser.add_argument('infile', metavar='NETWORK.py', help='Network creation file (in python) to upgrade')
    parser.add_argument('outfile', metavar='NEW.py', help='location of the result file')
    parser.add_argument('--type', metavar='TYPE', help='tool of resource to upgrade, one of: network, tool')
    return parser


def find_tool(toolspec):
    if isinstance(toolspec, str):
        namespace = None
        toolname = toolspec
        version = None
    elif isinstance(toolspec, tuple):
        if len(toolspec) == 2:
            toolname, version = toolspec
            namespace = None
        elif len(toolspec) == 3:
            namespace, toolname, version = toolspec
        else:
            raise exceptions.FastrToolUnknownError(f'Cannot use illegal tool spec {toolspec} with len {len(toolspec)}')
    else:
        raise exceptions.FastrToolUnknownError(
            f'Invalid toolspec {toolspec} type to find a tool, must be str or tuple, got a {type(toolspec).__name__}'
        )

    if version is None and '/' in toolname:
        toolname, version = toolname.split('/', 1)

    if namespace is None and '.' in toolname:
        namespace, toolname = toolname.rsplit('.', 1)

    if isinstance(version, str):
        version = Version(version)

    if namespace:
        namespace = '/'.join(namespace.split('.'))

    matching_tools = [t for t in fastr.tools.values() if t._id == toolname and
                      (version is None or t.command_version == version) and
                      (namespace is None or t.namespace == namespace)]

    if not matching_tools:
        raise exceptions.FastrToolUnknownError('Could not find tool: {}/{}:{}'.format(
            namespace or '$NAMESPACE',
            toolname,
            version or '$VERSION'
        ))

    matching_tools.sort(key=lambda x: (x.command_version, x.version), reverse=True)

    return matching_tools[0]


def upgrade_network(infile, outfile):
    try:
        from redbaron import RedBaron
    except ImportError:
        print("The upgrade command needs the redbaron package which is currently not installed. "
              "Please install the package.\n\nExample using pip:\n  $ pip install redbaron")
        return

    fastr.log.info('Loading input file...')
    with open(infile, 'r') as file_handle:
        code_file = RedBaron(file_handle.read())

    # Add import of resourcelimit
    fastr.log.info('Adding ResourceLimit import')
    for node in code_file.find_all('ImportNode', value=lambda x: x.dumps() == 'fastr'):
        node.insert_after('from fastr.api import ResourceLimit')

    # Replace all fastr.Network by fastr.create_network
    fastr.log.info('Updating network creation...')
    matches = code_file.find_all('AtomtrailersNode', value=lambda x: x.dumps().startswith('fastr.Network'))
    network_id = None
    for match in matches:
        if match.value[1].value != 'Network':
            continue

        match.value[1] = 'create_network'

        # Save the network id for later use
        for nr, subnode in enumerate(match.value[2]):
            if (subnode.target is None and nr == 0) or subnode.target.value == 'id_':
                try:
                    network_id = eval(subnode.value.dumps(), EVAL_GLOBALS, {})
                except:
                    network_id = subnode.value.dumps()

    # Change arguments for functions
    fastr.log.info('Updating method arguments...')
    for arg in code_file.find_all('CallArgumentNode'):
        if arg.target and arg.target.value == 'id_':
            arg.target.value = 'id'

        if arg.target and arg.target.value in ['nodegroup', 'sourcegroup']:
            arg.target.value = 'node_group'

        if arg.target and arg.target.value in ['stepid']:
            arg.target.value = 'step_id'

    # Add tool version and update tool name
    fastr.log.info('Updating create_node to include more version info...')
    for node in code_file.find_all('CallNode'):
        # Skip elements which do not call anything?
        if node.previous is None:
            continue

        if node.previous.value in ['create_source', 'create_constant', 'create_sink']:
            fastr.log.info(f'Updating {node.previous.value} statement: {node.parent}')
            if str(node.value[0].target) == 'datatype':
                node.value[0].target = ''

            try:
                data_type = eval(node.value[0].value.dumps(), EVAL_GLOBALS, {})
            except Exception:
                data_type = node.value[0].value

            node.value[0].value = f"'{data_type}'"

        elif node.previous.value == 'create_node':
            fastr.log.info(f'Updating create_node statement: {node.parent}')

            if str(node.value[0].target) == 'tool':
                node.value[0].target = ''

            tool_def = node.value[0].value.dumps()
            try:
                tool = eval(tool_def, EVAL_GLOBALS, {})
                tool = find_tool(tool)
                tool_version = tool.version
                tool_spec = tool.ns_id
                fastr.log.info(f'Found tool {tool}')
            except:
                fastr.log.error(f'Could not find tool based on: {tool_def}, using {tool_def} with tool_version="unknown"')
                tool_spec = tool_def
                tool_version = 'unknown'

            node.value[0].value = f"'{tool_spec}'"
            full_syntax = node.fst()
            if len(full_syntax['value']) < 3:
                raise ValueError('Cannot understand indentation of:\n{}'.format(node))

            # Get comma spec
            if len(full_syntax['value']) < 4:
                comma = deepcopy(full_syntax['value'][1])
            else:
                comma = deepcopy(full_syntax['value'][3])

            # Get call argument spec
            if not full_syntax['value'][2]['target']:
                full_syntax['value'][2]['target'] = {'type': 'name', 'value': 'id'}
            extra_arg = deepcopy(full_syntax['value'][2])

            if comma['type'] != 'comma' or extra_arg['type'] != 'call_argument':
                raise ValueError('Cannot understand syntax of:\n{}'.format(node))

            # Adapt call spec to correct values
            extra_arg['target']['value'] = 'tool_version'
            extra_arg['value']['value'] = f"'{tool_version}'"
            extra_arg['target']['type'] = 'name'

            # Insert and rebuild
            full_syntax['value'].insert(2, extra_arg)
            full_syntax['value'].insert(3, comma)
            node.replace(node.from_fst(full_syntax))

            # Get resource limits
            args = []
            to_remove = []
            for subnode in node.value:
                if subnode.target is not None and subnode.target.value == 'cores':
                    args.append(f'cores={subnode.value}')
                    to_remove.append(subnode)

                if subnode.target is not None and subnode.target.value == 'memory':
                    args.append(f'memory={subnode.value}')
                    to_remove.append(subnode)

                if subnode.target is not None and subnode.target.value == 'walltime':
                    args.append(f'time={subnode.value}')
                    to_remove.append(subnode)

            for subnode in to_remove:
                node.remove(subnode)

            if len(args) > 0:
                node.value.append('resources=ResourceLimit({})'.format(', '.join(args)))

    # Update draw_network -> draw and all argument changes
    fastr.log.info('Updating network draw to to new syntax...')
    for node in code_file.find_all('CallNode'):
        if node.previous.value != 'draw_network':
            continue

        node.previous.value = 'draw'

        name = network_id
        parsed_name = True
        img_format = 'svg'
        parsed_img_format = True
        file_path_node = None
        img_format_node = None
        for nr, subnode in enumerate(node.value):
            if (subnode.target is None and nr == 0) or (subnode.target is not None and subnode.target.value == 'name'):
                try:
                    name = eval(subnode.value.dumps(), EVAL_GLOBALS, {})
                except Exception:
                    name = subnode.value.dumps()
                    parsed_name = False

                file_path_node = subnode

            if (subnode.target is None and nr == 1) or (subnode.target is not None and subnode.target.value == 'img_format'):
                try:
                    img_format = eval(subnode.value.dumps(), EVAL_GLOBALS, {})
                except Exception:
                    img_format = subnode.value.dumps()
                    parsed_img_format = False

                img_format_node = subnode

            if (subnode.target is None and nr == 2) or (subnode.target is not None and subnode.target.value == 'draw_dimension'):
                subnode.target = 'draw_dimensions'

            if (subnode.target is None and nr == 3) or (subnode.target is not None and subnode.target.value == 'expand_macro'):
                subnode.target = 'expand_macros'

        # Construct correct file_path value
        if parsed_name and parsed_img_format:
            file_path_value = "'{}.{}'"
        elif parsed_name and not parsed_img_format:
            file_path_value = "'{}.' + {}"
        elif not parsed_name and parsed_img_format:
            file_path_value = "{} + '.{}'"
        else:
            file_path_value = "{} + '.' + {}"

        file_path_value = file_path_value.format(name, img_format)

        # Replace name and format arguments with the file_path argument
        if file_path_node is not None:
            file_path_node.replace("file_path={}".format(file_path_value))

        if img_format_node is not None:
            if file_path_node is not None:
                index = node.index_on_parent

                full_syntax = node.fst()
                full_syntax['value'].pop(node.index_on_parent)
                if len(full_syntax['value']) > index:
                    full_syntax['value'].pop(node.index_on_parent)

                node.replace(node.from_fst(full_syntax))
            else:
                img_format_node.replace("file_path={}".format(file_path_value))

    fastr.log.info('Writing output file...')
    with open(outfile, 'w') as output_file_handle:
        output_file_handle.write(code_file.dumps())


def upgrade_tool(infile, outfile):
    from fastr.core.tool import Tool

    fastr.log.info(f'Loading tool definition from {infile}')
    tool = Tool.loadf(infile)
    fastr.log.info(f'Saving new tool definition to {outfile}')
    tool.dumpf(outfile, method='yaml')


def main():
    """
    Upgrade a fastr 2.x python file to fastr 3.x syntax
    """
    # No arguments were parsed yet, parse them now
    parser = add_parser_doc_link(get_parser(), __file__)
    args = parser.parse_args()

    input_file = args.infile
    output_file = args.outfile
    resource_type = args.type

    if resource_type is None:
        if input_file.endswith('.py'):
            resource_type = 'network'
        elif input_file.endswith('.xml'):
            resource_type = 'tool'

    if resource_type == 'network':
        try:
            upgrade_network(args.infile, args.outfile)
        except exceptions.FastrToolUnknownError as exception:
            fastr.log.error('Encountered exception: {}'.format(exception))
    elif resource_type == 'tool':
        upgrade_tool(input_file, output_file)
    else:
        fastr.log.error(f'Invalid resource type, should be "network" or "tool", found {resource_type}')


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module keeps track of the version of the currently used Fastr framework.
It can check its version from mercurial or a saved file
"""
import os
import re
from typing import Optional, Tuple

try:
    import git
    _GIT_PRESENT = True
except ImportError:
    _GIT_PRESENT = False

_VERSIONFILE = os.path.join(os.path.dirname(__file__), 'versioninfo')
_FASTR_SRC_DIR = os.path.normpath(os.path.join(os.path.dirname(__file__), '..'))


def get_base_version() -> Optional[str]:
    """
    Get the version from the top-level version file

    :return: the version
    :rtype str:
    """
    try:
        with open(os.path.join(os.path.dirname(__file__), '..', 'version'), 'r') as version_file:
            return version_file.read().strip()
    except IOError:
        return None


def get_git_info() -> Tuple[Optional[str], Optional[str]]:
    if not _GIT_PRESENT:
        return None, None

    try:
        search_path = os.path.dirname(os.path.abspath(__file__))
        repo = git.Repo(path=search_path, search_parent_directories=True)
    except git.InvalidGitRepositoryError:
        return None, None

    git_head, git_branch = repo.head.object.name_rev.split(" ", 1)
    return git_head, git_branch


def get_saved_version() -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """
    Get cached version from file

    :return: tuple with version, head revision and branch
    """
    try:
        with open(_VERSIONFILE, 'r') as version_file:
            data = version_file.read().split()
            data = dict(x.split('=', 1) for x in data)

        return data.get('version', None), data.get('git_head', None), data.get('git_branch', None)
    except IOError:
        return None, None, None


def save_version(current_version: str, current_hg_head: str, current_hg_branch: str):
    """
    Cache the version information (useful for when installing)

    :param str current_version: version
    :param str current_hg_head: mercurial head revision
    :param str current_hg_branch: mercurial branch
    :return:
    """
    with open(_VERSIONFILE, 'w') as version_file:
        version_file.write('version={}\n'.format(current_version))
        version_file.write('git_head={}\n'.format(current_hg_head))
        version_file.write('git_branch={}\n'.format(current_hg_branch))


def clear_version():
    """
    Remove the cached version info
    """
    os.remove(_VERSIONFILE)


# Try to get version info
version = get_base_version()
git_head, git_branch = get_git_info()

# Note that we imported from a mercurial repository
from_git = git_branch is not None

# Try to get missing info from saved versioninfo
if version is None or git_head is None or git_branch is None:
    _n_version, _n_hg_head, _n_hg_branch = get_saved_version()

    version = version or _n_version
    git_head = git_head or _n_hg_head
    git_branch = git_branch or _n_hg_branch

git_head = git_head or 'unknown'
git_branch = git_branch or 'unknown'

full_version = '{}_{}_{}'.format(version, git_branch, git_head[:8])
not_master_branch = re.match(f'^tags/{version}(\^\d+)?$', git_branch) is None


# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import sys
import urllib.parse as up

import fastr
from fastr import exceptions, resources
from fastr.data import url as urltools
from fastr.utils.cmd import add_parser_doc_link


def get_parser():
    parser = argparse.ArgumentParser(description="executes an ioplugin",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-i', '--input', nargs='+', type=str, required=True,
                        help="The url to process (can also be a list)")
    parser.add_argument('-o', '--output', nargs='+', type=str, metavar='OUTPUT', required=True,
                        help="The output urls in vfs scheme (can also be a"
                             " list and should be the same size as --inurl)")
    parser.add_argument('-d', '--datatype', nargs='+', default=None, type=str,
                        help="The datatype of the source/sink data to handle")
    return parser


def sink():
    # No arguments were parsed yet, parse them now
    parser = add_parser_doc_link(get_parser(), __file__)
    args = parser.parse_args()

    fastr.log.info('Using Python version {} loaded from {}'.format(sys.version, sys.executable))
    fastr.log.info(('Arguments:\ninput: {}\noutput: {}\ndatatype: {}').format(
        args.input,
        args.output,
        args.datatype
    ))

    inputs = args.input
    outputs = args.output
    datatypes = args.datatype

    if len(datatypes) == 1:
        datatypes *= len(inputs)

    for input_, output, datatype in zip(inputs, outputs, datatypes):
        fastr.log.info('Processing {} --> {} [{}]'.format(input_, output, datatype))

        if datatype is not None:
            datatype = fastr.types[datatype]

        fastr.log.debug("datatype to use: {}".format(datatype))

        if urltools.isurl(input_):
            message = 'For Sink behaviour the sink should be a path, not a URL! (found {})'.format(input_)
            fastr.log.critical(message)
            raise exceptions.FastrValueError(message)

        parsed_output = up.urlparse(output)
        try:
            plugin = resources.ioplugins[parsed_output.scheme]
        except KeyError:
            message = "No valid scheme is supplied in: {} (Found {}). The following schemes are supported: {}".format(output, parsed_output.scheme, ' '.join(resources.ioplugins.keys()))
            fastr.log.error(message)
            raise exceptions.FastrUnknownURLSchemeError(message)

        plugin.push_sink_data(input_, output, datatype)

    # Cleanup all plugins
    resources.ioplugins.cleanup()


def main():
    """
    Command line access to the IOPlugin sink
    """
    try:
        sink()
    except Exception as exception:
        # Pass error on to main process and raise error again
        exception = (
            type(exception).__name__,
            str(exception),
            exception.filename if hasattr(exception, 'filename') else None,
            exception.linenumber if hasattr(exception, 'linenumber') else None,
        )

        print('__FASTR_ERRORS__ = {}'.format(
            json.dumps(
                [exception]
            )
        ))
        raise


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A module implenting a lock that ensures a directory is only being used
by a single fastr run.
"""

import os
from pathlib import Path
import psutil
from typing import Optional, Union

from ..exceptions import FastrLockNotAcquired


class DirectoryLock:
    """
    A lock for a directory, it creates a directory to set the locked state and
    if successful writes the pid in a file inside that directory to claim the
    lock
    """
    lock_dir_name: str = '.fastr.lock'
    pid_file_name: str = 'pid'

    def __init__(self, directory: Union[str, Path]):
        self._directory = Path(directory)
        self._acquired = False

    @property
    def lock_dir(self) -> Path:
        return self._directory / self.lock_dir_name

    @property
    def pid_file(self) -> Path:
        return self.lock_dir / self.pid_file_name

    def get_pid(self) -> Optional[int]:
        try:
            lock_pid = int(self.pid_file.read_text())
        except (FileNotFoundError, ValueError):
            lock_pid = None

        return lock_pid

    @staticmethod
    def _checkpid(pid) -> bool:
        return psutil.pid_exists(pid)

    def acquire(self) -> bool:
        # If acquired, validate lock is still in place
        if self._acquired:
            lock_pid = self.get_pid()

            if lock_pid is None:
                self.release()
                return False

            if lock_pid == os.getpid():
                return True
            else:
                self.release()
                return False

        # Try to create a lock directory and make sure it does not exist
        if self.lock_dir.exists() or self.lock_dir.is_dir():
            if self.pid_file.exists():
                pid = self.get_pid()

                if not self._checkpid(pid):
                    self.release(force=True)
                else:
                    return False
            else:
                return False

        try:
            self.lock_dir.mkdir(parents=False, exist_ok=False)
        except FileExistsError:
            return False

        # Register creating PID in file in lock dir
        self.pid_file.write_text(str(os.getpid()))

        self._acquired = True

        return True

    def release(self, force: bool=False):
        if not force and not self._acquired:
            return

        lock_pid = self.get_pid()

        if lock_pid is None or force or lock_pid == os.getpid():
            try:
                self.pid_file.unlink()
            except FileNotFoundError:
                pass

            try:
                self.lock_dir.rmdir()
            except FileNotFoundError:
                pass

        self._acquired = False

    def __enter__(self):
        if not self.acquire():
            raise FastrLockNotAcquired(self._directory)

    def __exit__(self, type, value, traceback):
        self.release()

    def __del__(self):
        self.release()





from abc import abstractmethod
import re
from typing import Optional, Sequence

from sympy.core.expr import Expr
from sympy.core.symbol import Symbol
from sympy import Add, Mul

from .samples import SampleIndex
from .. import exceptions
from ..helpers import log
from ..execution import inputoutputrun


class CardinalitySpec:
    def __init__(self, parent):
        self.parent = parent

    @abstractmethod
    def __str__(self) -> str:
        """
        String version of the cardinality spec, should be parseable by
        create_cardinality
        """

    def __repr__(self) -> str:
        """
        Console representation of the cardinality spec
        """
        return '<{} {}>'.format(type(self).__name__, self)

    @abstractmethod
    def __eq__(self, other: 'CardinalitySpec') -> bool:
        """Test for equality"""

    def __ne__(self, other: 'CardinalitySpec') -> bool:
        return not self == other

    @property
    def predefined(self):
        """
        Indicate whether the cardinality is predefined or can only be calculated
        after execution
        """
        return False

    def validate(self,
                 payload: Optional[dict],
                 cardinality: int,
                 planning: bool = True) -> bool:
        """
        Validate cardinality given a payload and cardinality

        :param payload: Payload of the corresponding job
        :param cardinality: Cardinality to validate
        :param planning: Indicate whether the is for the planning phase or not
        :return: Validity of the cardinality given the spec and payload
        """
        if isinstance(cardinality, (Symbol, Expr)):
            return planning

        result = self._validate(payload, cardinality)
        if isinstance(result, bool):
            return result
        elif isinstance(result, (Symbol, Expr)):
            return True
        else:
            raise exceptions.FastrTypeError(
                'Cardinality validation should be either a sympy expression'
                ' or boolean, found a {}'.format(type(result).__name__)
            )

    @abstractmethod
    def _validate(self,
                  payload: Optional[dict],
                  cardinality: int) -> bool:
        """
        Validate cardinality given a payload and cardinality

        :param payload: Payload of the corresponding job
        :param cardinality: Cardinality to validate
        :return: Validity of the cardinality given the spec and payload
        """

    def calculate_planning_cardinality(self) -> Optional[int]:
        """
        Calculate the cardinality given the node and spec, for cardinalities
        that only have validation and not a pre-calculable value, this return
        None.
        :return: calculated cardinality
        """
        return None

    def calculate_execution_cardinality(self, key=None) -> Optional[int]:
        """
        Calculate the cardinality given the node and spec, during execution
        this should be available and not give unknowns once the data is present
        and the key is given.

        :param key: Key for which the cardinality is calculated
        :return: calculated cardinality
        """
        if key is None:
            return None

        sample = self.parent.samples.get(key)

        if sample is None:
            return None

        return sample.cardinality

    def calculate_job_cardinality(self, payload: dict) -> Optional[int]:
        """
        Calculate the actually cardinality when a job needs to know how many
        arguments to create for a non-automatic output.
        """
        return None


class IntCardinalitySpec(CardinalitySpec):
    def __init__(self, parent, value: int):
        super().__init__(parent)
        self.value = value

    def __eq__(self, other) -> bool:
        """Test for equality"""
        if type(self) != type(other):
            return NotImplemented

        return self.value == other.value

    def __str__(self) -> str:
        return '{}'.format(self.value)

    def calculate_job_cardinality(self, payload: dict) -> Optional[int]:
        return self.value

    @property
    def predefined(self):
        return True

    def _validate(self, payload: dict, cardinality: int) -> bool:
        return self.value == cardinality

    def calculate_planning_cardinality(self) -> int:
        return self.value

    def calculate_execution_cardinality(self, key=None) -> int:
        return self.value


class MinCardinalitySpec(CardinalitySpec):
    def __init__(self, parent, value: int):
        super().__init__(parent)
        self.value = value

    def __eq__(self, other):
        """Test for equality"""
        if type(self) != type(other):
            return NotImplemented
        return self.value == other.value

    def __str__(self) -> str:
        return '{}-*'.format(self.value)

    def _validate(self, payload: dict, cardinality: int) -> bool:
        return cardinality >= self.value


class MaxCardinalitySpec(CardinalitySpec):
    def __init__(self, parent, value: int):
        super().__init__(parent)
        self.value = value

    def __eq__(self, other):
        """Test for equality"""
        if type(self) != type(other):
            return NotImplemented
        return self.value == other.value

    def __str__(self) -> str:
        return '*-{}'.format(self.value)

    def _validate(self, payload: dict, cardinality: int) -> bool:
        return cardinality <= self.value


class RangeCardinalitySpec(CardinalitySpec):
    def __init__(self, parent, min: int, max: int):
        super().__init__(parent)
        self.min = min
        self.max = max

    def __eq__(self, other):
        """Test for equality"""
        if type(self) != type(other):
            return NotImplemented
        return self.min == other.min and self.max == other.max

    def __str__(self) -> str:
        return '{}-{}'.format(self.min, self.max)

    def _validate(self, payload: dict, cardinality: int) -> bool:
        return self.min <= cardinality <= self.max


class ChoiceCardinalitySpec(CardinalitySpec):
    def __init__(self, parent, options: Sequence[int]):
        super().__init__(parent)
        self.options = tuple(options)

    def __eq__(self, other):
        """Test for equality"""
        if type(self) != type(other):
            return NotImplemented
        return self.options == other.options

    def __str__(self) -> str:
        return '[{}]'.format(','.join(str(x) for x in self.options))

    def _validate(self, payload: dict, cardinality: int) -> bool:
        return cardinality in self.options


class AnyCardinalitySpec(CardinalitySpec):
    def _validate(self, payload: dict, cardinality: int) -> bool:
        return True

    def __eq__(self, other):
        """Test for equality"""
        if type(self) != type(other):
            return NotImplemented
        return True

    def __str__(self) -> str:
        return 'any'


class AsCardinalitySpec(CardinalitySpec):
    def __init__(self, parent, target):
        super().__init__(parent)
        self.target = target

    def __eq__(self, other):
        """Test for equality"""
        if type(self) != type(other):
            return NotImplemented

        return self.target == other.target

    def __str__(self) -> str:
        return 'as:{}'.format(self.target)

    @property
    def predefined(self):
        return True

    @property
    def node(self):
        return self.parent.node

    def _validate(self, payload: dict, cardinality: int) -> bool:
        if payload is None:
            return True

        if self.target.endswith(']'):
            # Need to index an input or output to get the cardinality
            start = self.target.find('[')
            index = int(self.target[start + 1:-1])
            target = self.get_target()
            value = payload['inputs'].get(target)
            if value is None:
                value = payload['outputs'].get(target, [])

            if len(value) == 0:
                # Argument not given, thus expected cardinality is zero
                expected_cardinality = 0
            else:
                try:
                    index_key = list(value.keys())[index]
                except IndexError:
                    length = len(list(value.keys()))
                    raise exceptions.FastrIndexError(
                        'Cardinality references to invalid field '
                        '(target {} has length {} '.format(target, length) +
                        'and thus does not have index {})'.format(index)
                    )

                expected_cardinality = len(value[index_key])

        else:
            value = payload['inputs'].get(self.target)
            if value is None:
                value = payload['outputs'].get(self.target, [])
            expected_cardinality = len(value)

        return cardinality == expected_cardinality

    def get_target(self) -> str:
        # Check if we need to index the target
        if self.target.endswith(']'):
            # Need to get length of the as input
            start = self.target.find('[')
            target = self.target[:start]
        else:
            target = self.target

        return target

    def get_ordereddict_cardinality(self):
        # Get the index of the OrderedDict target
        start = self.target.find('[')
        index = int(self.target[start + 1:-1])

        # Get the full target and cardinality
        target = self.node.inputs.get(self.get_target())
        cardinality = target.cardinality()

        # On Job execution, target is inputoutputrun
        if isinstance(target, inputoutputrun.InputRun):
            # Multiple subinputs, index. Occurs in job execution.
            cardinality = target.get_subinput_cardinality(index)

        elif isinstance(cardinality, Add):
            # Multiple elements added, split. Occurs in planning.
            elements = cardinality.args
            n_elements = len(elements)
            if index > n_elements:
                raise exceptions.FastrIndexError(
                    'Cardinality references to invalid field '
                    '(target {} has length {} '.format(target, n_elements) +
                    'and thus does not have index {})'.format(index)
                )
            else:
                cardinality = elements[index]

        elif isinstance(cardinality, Mul):
            # All elements are multplied: del multiplier, as we
            # only request a single element. Occurs in execution.
            elements = cardinality.args
            elements = list(elements)
            del elements[0]  # Delete the multiplier
            cardinality = cardinality.func(*elements)

        return cardinality

    def calculate_planning_cardinality(self) -> Optional[int]:
        # Get cardinality from target if possible
        check_target = self.node.inputs.get(self.get_target())
        if check_target is not None:
            if self.target.endswith(']'):
                # print('Planning: ' + str(self.target) + str(check_target.cardinality()))
                return self.get_ordereddict_cardinality()
            else:
                return self.node.inputs.get(self.target).cardinality()
        else:
            raise exceptions.FastrCardinalityError(
                'Cardinality references to invalid field '
                '({} is not an Input in this Node)'.format(self.target)
            )

    def calculate_execution_cardinality(self, key=None) -> Optional[int]:
        check_target = self.node.inputs.get(self.get_target())
        if check_target is None:
            raise exceptions.FastrCardinalityError('Cardinality references to invalid field '
                                                   '({} is not an Input in this Node)'.format(self.target))

        if self.target.endswith(']'):
            # Target is element of OrderedDict
            return self.get_ordereddict_cardinality()
        else:
            # Normal procedure
            target = check_target

        if key is None:
            # No key is used, call target without key
            cardinality = target.cardinality(None)
        elif all(x == 0 for x in target.size):
            # Target is empty, cardinality can be set to 0
            cardinality = 0
        elif target.size == (1,):
            # Target has only sample, it will be repeated, use first sample
            cardinality = target.cardinality((0,))
        elif len(self.node.input_groups) == 1:
            # The InputGroups are not mixed, we can request the key
            if len(key) == len(target.size):
                cardinality = target.cardinality(key)
            else:
                index_map = dict(zip(self.parent.dimnames, key))
                lookup = {v: dimname for dimname in self.parent.dimnames for value in self.node.parent.nodegroups.values() if dimname in value for v in value}
                lookup.update({x: x for x in self.parent.dimnames})
                if all(x in lookup for x in target.dimnames):
                    # Print there is broadcasting going on, we need to undo that here
                    matched_dimnames = [lookup[x] for x in target.dimnames]
                    matched_index = SampleIndex(index_map[x] for x in matched_dimnames)
                    cardinality = target.cardinality(matched_index)
                else:
                    raise exceptions.FastrSizeMismatchError(
                        'InputGroup has inconsistent size/dimension info for Input '
                        '{}, cannot figure out broadcasting used!'.format(target.fullid)
                    )
        else:
            log.debug('Unmixing key "{}" for cardinality retrieval'.format(key))
            # The InputGroups are mixed, find the part of the ID relevant to this Input
            test = self.node.input_group_combiner.unmerge(key)
            index = test[target.input_group]

            if len(index) == len(target.size):
                cardinality = target.cardinality(index)
            else:
                raise exceptions.FastrSizeMismatchError('TODO: add broadcasting to this branch?')

        return cardinality

    def calculate_job_cardinality(self, payload: dict) -> Optional[int]:
        target = payload['inputs'].get(self.target)

        if target is None:
            raise exceptions.FastrCardinalityError(
                'Cardinality references to invalid field '
                '({} is not an Input in this Node)'.format(self.target)
            )

        return len(target)


class ValueCardinalitySpec(CardinalitySpec):
    def __init__(self, parent, target):
        super().__init__(parent)
        self.target = target

    def __eq__(self, other):
        """Test for equality"""
        if type(self) != type(other):
            return NotImplemented

        return self.target == other.target

    def __str__(self) -> str:
        return 'val:{}'.format(self.target)

    @property
    def node(self):
        return self.parent.node

    def _validate(self, payload: dict, cardinality: int) -> bool:
        if payload is None:
            return True

        value = payload['inputs'].get(self.target)
        if value is None:
            value = payload['outputs'].get(self.target)

        if value is None:
            raise exceptions.FastrValueError('Cannot calculate val: type cardinality if value not in payload!')

        if len(value) != 1:
            return False

        try:
            value = int(value[0].value)
        except (ValueError, TypeError):
            return False

        return cardinality == value

    def calculate_execution_cardinality(self, key=None) -> Optional[int]:
        if self.target in self.node.inputs:
            # We cannot access to the jobs inputs it appears, so we
            # check if the output has already been generated.
            if self.parent.samples is not None and key in self.parent.samples:
                value = self.parent.samples[key].data
                log.debug('Got val via output data result, got {}'.format(value))
                return len(value)
            else:
                log.debug('Cannot get val: cardinality if there is no execution data!')
                return None
        elif self.target in self.node.outputs:
            # Get the value an output
            if key is None:
                return None

            output = self.node.outputs[self.target]

            if output.samples is None:
                return None

            # Try to cast via str to int (To make sure Int datatypes fares well)
            try:
                return int(str(output[key]))
            except exceptions.FastrKeyError:
                return None

    def calculate_job_cardinality(self, payload: dict) -> Optional[int]:
        target = payload['inputs'].get(self.target)

        if target is None:
            raise exceptions.FastrCardinalityError(
                'Cardinality references to invalid field '
                '({} is not an Input in this Node)'.format(self.target)
            )

        if len(target) != 1:
            raise exceptions.FastrValueError(
                'Cannot determine cardinality from multiple values '
                '(requested {}, found  {} as value)'.format(self, target)
            )

        return int(str(target[0]))


def create_cardinality(desc: str, parent) -> CardinalitySpec:
    """
    Create simplified description of the cardinality. This changes the
    string representation to a tuple that is easier to check at a later
    time.

    :param desc: the string version of the cardinality
    :param parent: the parent input or output to which this cardinality spec belongs
    :return: the simplified cardinality description
    :raises FastrCardinalityError: if the Input/Output has an incorrect
            cardinality description.

    The translation works with the following table:

    ==================== ============================= ===============================================================
    cardinality string   cardinality spec              description
    ==================== ============================= ===============================================================
    ``"*"``, ``any``     ``('any',)                    Any cardinality is allowed
    ``"N"``              ``('int', N)``                A cardinality of N is required
    ``"N-M"``            ``('range', N, M)``           A cardinality between N and M is required
    ``"*-M"``            ``('max', M)``                A cardinality of maximal M is required
    ``"N-*"``            ``('min', N)``                A cardinality of minimal N is required
    ``"[M,N,...,O,P]"``  ``('choice', [M,N,...,O,P])`` The cardinality should one of the given options
    ``"as:input_id"``    ``('as', 'input_id')``        The cardinality should match the cardinality of the given Input
    ``"val:input_id"``   ``('val', 'input_id')``       The cardinliaty should match the value of the given Input
    ==================== ============================= ===============================================================

    .. note::

        The maximumu, minimum and range are inclusive

    """

    if isinstance(desc, int) or re.match(r'^\d+$', desc) is not None:
        # N
        return IntCardinalitySpec(parent, int(desc))
    elif desc in ['*', 'any', 'unknown']:
        # * (anything is okay)
        return AnyCardinalitySpec(parent)
    elif re.match(r'^\[\d+(,\d+)*]', desc) is not None:
        # [M,N,..,O,P]
        return ChoiceCardinalitySpec(parent, [int(x) for x in desc[1:-1].split(',')])
    elif '-' in desc:
        match = re.match(r'^(\d+|\*)-(\d+|\*)$', desc)
        if match is None:
            raise exceptions.FastrCardinalityError("Not a valid cardinality description string (" + desc + ")")

        lower, upper = match.groups()

        if lower == '*' and upper == '*':
            # *-* (anything is okay)
            return AnyCardinalitySpec(parent)
        elif lower == '*' and upper != '*':
            # N-*
            return MaxCardinalitySpec(parent, int(upper))
        elif lower != '*' and upper == '*':
            # *-M
            return MinCardinalitySpec(parent, int(lower))
        else:
            # N-M
            return RangeCardinalitySpec(parent, int(lower), int(upper))

    elif desc.startswith("as:"):
        # as:other
        return AsCardinalitySpec(parent, desc[3:])
    elif desc.startswith("val:"):
        # val:other
        return ValueCardinalitySpec(parent, desc[4:])
    else:
        raise exceptions.FastrCardinalityError("Not a valid cardinality description string (" + desc + ")")



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Manager class for Plugins in the fastr system
"""

import collections.abc

from ...abc.baseplugin import BasePlugin, Plugin
from ...abc.basepluginmanager import BasePluginManager
from ...helpers import config, log


class PluginsView(collections.abc.MutableMapping):
    """
    A collection that acts like view of the plugins of another plugin manager.
    This is a proxy object that only gives access the plugins of a certain
    plugin class. It behaves like a mapping and is used as the data object for
    a PluginSubManager.
    """
    def __init__(self, parent, plugin_class):
        """
        Constructor for the plugins view

        :param BasePluginManager parent: the parent plugin manager
        :param class plugin_class: the class of the plugins to expose
        """
        self.plugin_class = plugin_class
        self.parent = parent

    def filter_plugin(self, plugin):
        if self.plugin_class.instantiate and isinstance(plugin, self.plugin_class):
            return True
        elif not self.plugin_class.instantiate and (isinstance(plugin, type) and issubclass(plugin, self.plugin_class)):
            return True
        else:
            return False

    def __getitem__(self, item):
        result = self.parent[item]

        if not self.filter_plugin(plugin=result):
            raise KeyError(item)

        return result

    def __setitem__(self, key, value):
        if not self.filter_plugin(value):
            raise TypeError(value)

        if key in self.parent and not self.filter_plugin(self.parent[key]):
            raise TypeError(key)

        self.parent[key] = value

    def __delitem__(self, key):
        if key in self.parent and not self.filter_plugin(self.parent[key]):
            raise TypeError(key)

        del self.parent[key]

    def __len__(self):
        return sum(1 for v in self.parent.values() if self.filter_plugin(v))

    def __iter__(self):
        for key, value in self.parent.items():
            if not self.filter_plugin(value):
                continue

            yield key


class PluginSubManager(BasePluginManager):
    """
    A PluginManager that is a selection of a parent plugin manger. It uses the
    PluginsView to only exponse part of the parent PluginManager. This is used
    to create plugin plugins.managers for only certain types of plugins (e.g. IOPlugins)
    without loading them multiple times.
    """
    def __init__(self, parent, plugin_class):
        self.parent = parent
        self._plugin_class = plugin_class
        self._data_link = PluginsView(parent=parent, plugin_class=plugin_class)
        super().__init__()

    @property
    def data(self):
        return self._data_link

    @property
    def plugin_class(self):
        """
        PluginSubManagers only expose the plugins of a certain class
        """
        return self._plugin_class


class PluginManager(BasePluginManager):
    def __init__(self, path=None):
        if path is None:
            path = config.plugins_path

        super().__init__(path=path, recursive=True)

    @property
    def plugin_class(self):
        """
        The plugin manager contains any Plugin subclass
        """
        return Plugin

    def __setitem__(self, key, value):
        """
        Store an item in the BaseManager, will ignore the item if the key is
        already present in the BaseManager.

        :param name: the key of the item to save
        :param value: the value of the item to save
        :return: None
        """
        if not (isinstance(value, Plugin) or issubclass(value, Plugin)):
            raise TypeError(value)

        super(PluginManager, self).__setitem__(key, value)

    def _store_item(self, name, value):
        """
        Store an item in the BaseManager, will ignore the item if the key is
        already present in the BaseManager.

        :param name: the key of the item to save
        :param value: the value of the item to save
        :return: None
        """
        if name in list(self.keys()):
            log.warning('Skipping {} from {} (the plugin is already in the {})'.format(name, value.filename, type(self).__name__))
        else:

            # Set the module to the fastr plugins
            if isinstance(value, BasePlugin):
                type(value).__module__ = 'fastr.plugins'
                setattr(self._module, value.id, type(value))
            else:
                value.__module__ = 'fastr.plugins'
                setattr(self._module, value.id, value)

            self[name] = value



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the S3Storage plugin for fastr
"""
import urllib.parse

import os

import fastr
import fastr.data
from fastr import exceptions
from fastr.core.ioplugin import IOPlugin

boto_loaded = False
try:
    import boto3
    import botocore
    boto_loaded = True
except ImportError:
    pass


class S3Filesystem(IOPlugin):
    """
    .. warning::

        As this IOPlugin is under development, it has not been thoroughly
        tested.

        example url: s3://bucket.server/path/to/resource
    """
    scheme = ('s3', 's3list')

    def __init__(self):
        # initialize the instance and register the scheme
        super(S3Filesystem, self).__init__()
        self._s3clients = {}

    def cleanup(self):
        pass

    @classmethod
    def test(cls):
        if not boto_loaded:
            raise exceptions.FastrImportError('Could not import the required boto3/botocore modules for this plugin')

    def expand_url(self, url):
        """
                Expand an S3 URL. This allows a source to collect multiple
                samples from a single url.

                :param str url: url to expand
                :return: the resulting url(s), a tuple if multiple, otherwise a str
                :rtype: str or tuple of str
        """
        if fastr.data.url.get_url_scheme(url) != 's3list':
            raise exceptions.FastrValueError('URL not of {} type!'.format('s3list'))

        try:
            data = self._get_file_as_var(url)
            return tuple((None, x.strip()) for x in data.strip().split('\n'))
        except botocore.exceptions.ClientError as e:
            raise self._translate_boto_to_fastr_exception(e)

    def fetch_url(self, inurl, outpath):
        """
        Get the file(s) or values from s3.

        :param inurl: url to the item in the data store
        :param outpath: path where to store the fetch data locally
        """
        if fastr.data.url.get_url_scheme(inurl) != 's3':
            raise exceptions.FastrValueError('URL not of {} type!'.format('s3'))

        parsed_url = self._parse_url(inurl)
        s3 = self._create_client(profile=parsed_url["profile"], server=parsed_url["server"], bucket=parsed_url["bucket"])

        try:
            s3.download_file(parsed_url['key'], outpath)
        except botocore.exceptions.ClientError as e:
            raise self._translate_boto_to_fastr_exception(e)

    def put_url(self, inpath, outurl):
        """
        Upload the files to the S3 storage

        :param inpath: path to the local data
        :param outurl: url to where to store the data in the external data store.
        """

        if fastr.data.url.get_url_scheme(outurl) != 's3':
            raise exceptions.FastrValueError('URL not of {} type!'.format('s3'))

        if os.path.isdir(inpath):
            # we have a directory
            try:
                listdir = os.listdir(inpath)
            except:
                # If no read access is available on a directory this exception gets thrown.
                raise exceptions.FastrIOError("Could not listdir {}".format(inpath))
            else:
                for path in listdir:
                    if not outurl[-1:] == '/':
                        outurl = outurl + '/'
                    # recursively add files to s3
                    self.put_url(os.path.join(inpath, path), outurl + path)
        else:
            # we have a file
            parsed_url = self._parse_url(outurl)
            s3 = self._create_client(profile=parsed_url["profile"], server=parsed_url["server"], bucket=parsed_url["bucket"])
            try:
                s3.upload_file(inpath, parsed_url["key"])
                return True # we need to return something, else ioplugin.py:push_sink_data generates a FastrIOError('Could not store data from {} to {}')
            except botocore.exceptions.ClientError as e:
                raise self._translate_boto_to_fastr_exception(e)

    def put_value(self, value, outurl):
        """
        Put the value in S3

        :param value: value to store
        :param outurl: url to where to store the data, starts with ``file://``
        """
        if fastr.data.url.get_url_scheme(outurl) != 's3':
            raise exceptions.FastrValueError('URL not of {} type!'.format('s3'))

        parsed_url = self._parse_url(outurl)
        s3 = self._create_client(profile=parsed_url["profile"], server=parsed_url["server"], bucket=parsed_url["bucket"])
        try:
            s3.put_object(Body=value, Key=parsed_url["key"])
            return True # we need to return something, else ioplugin.py:push_sink_data generates a FastrIOError('Could not store data from {} to {}')
        except botocore.exceptions.ClientError as e:
            raise self._translate_boto_to_fastr_exception(e)

    def fetch_value(self, inurl):
        """
        Fetch a value from S3

        :param inurl: url of the value to read
        :return: the fetched value
        """
        if fastr.data.url.get_url_scheme(inurl) != 's3':
            raise exceptions.FastrValueError('URL not of {} type!'.format('s3'))

        try:
            return self._get_file_as_var(inurl)
        except botocore.exceptions.ClientError as e:
            raise self._translate_boto_to_fastr_exception(e)

    def _translate_boto_to_fastr_exception(self, exception):
        res = exception.response
        if res['Error']['Code'] == "404":
            return exceptions.FastrDataTypeValueError("The file was not found: ")
        else:
            return exceptions.FastrIOError("{} {} {}".format(
                res['Error']['Code'],
                res['ResponseMetadata']['RequestId'],
                res['Error']['Message']
            ))

    def _get_file_as_var(self, url):
        parsed_url = self._parse_url(url)
        s3 = self._create_client(profile=parsed_url["profile"], server=parsed_url["server"], bucket=parsed_url["bucket"])

        try:
            return s3.Object(parsed_url['key']).get()['Body'].read().decode('utf-8')
        except botocore.exceptions.ClientError as e:
            raise

    def _parse_url(self, url):
        url = urllib.parse.urlparse(url)
        parts = url.netloc.split('@', 1)
        if len(parts) == 2:
            profile = parts[0]
            parts = parts[1].split('.', 1)
        else:
            profile = 'default'
            parts = parts[0].split('.', 1)

        if url.path[0] == '/':
            key = url.path[1:]
        else:
            key = url.path

        return {"bucket": parts[0], "server": parts[1], "profile": profile, "key": key}

    def _create_client(self, profile, server, bucket):
        client_key = "{}.{}".format(profile, server)
        if client_key in self._s3clients:
            s3 = self._s3clients[client_key]
        else:
            session = boto3.Session(profile_name=profile)
            s3 = session.resource('s3', endpoint_url='http://{}'.format(server)).Bucket(bucket)
            self._s3clients[client_key] = s3

        return s3



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Manager class for Plugins in the fastr system
"""

import sys
from types import ModuleType


class LazyModule(ModuleType):
    """
    A module that allows content to be loaded lazily from plugins. It generally
    is (almost) empty and gets (partially) populated when an attribute cannot be
    found. This allows lazy loading and plugins depending on other plugins.
    """
    def __init__(self, name, parent, plugin_manager):
        super(LazyModule, self).__init__(name)
        self._plugin_manager = plugin_manager
        self.__dict__.update(vars(parent))
        self._parent = parent
        sys.modules[parent.__name__] = self

    def __repr__(self):
        return super(LazyModule, self).__repr__().replace('module', 'lazy_module', 1)

    def __getattr__(self, item):
        """
        The getattr is called when getattribute does not return a value and is
        used as a fallback. In this case we try to find the value normally and
        will trigger the plugin manager if it cannot be found.

        :param str item: attribute to retrieve
        :return: the requested attribute
        """
        try:
            return super(LazyModule, self).__getattribute__(item)
        except AttributeError as exception:
            try:
                return self._plugin_manager[item]
            except KeyError:
                raise exception



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABCMeta, abstractmethod
from typing import Tuple, TypeVar

from sympy.core.symbol import Symbol

from .. import exceptions

SizeType = TypeVar('SizeType', int, Symbol)


class Dimension(object):
    """
    A class representing a dimension. It contains the name and size of the
    dimension.
    """
    __slots__ = ('_name', '_size')

    def __init__(self, name: str, size: SizeType):
        """
        The constructor for the dimension.

        :param str name: Name of the dimension
        :param size: Size fo the dimension
        :type size: int or Symbol
        """
        if not isinstance(name, str):
            raise exceptions.FastrTypeError("Dimension.name should be a str, "
                                            "found [{}] {}".format(type(name).__name__,
                                                                   size))

        self._name = name
        self._size = None
        self.size = size

    def __repr__(self) -> str:
        """
        String representation of a Dimension
        """
        return "<Dimension {d.name} ({d.size})>".format(d=self)

    def __eq__(self, other: 'Dimension') -> bool:
        """
        Dimension is the same if the name and size are the same
        """
        return self.name == other.name and self.size == other.size

    def __ne__(self, other: 'Dimension') -> bool:
        """
        The not equal test is simply the inverse of the equal test
        """
        return not self == other

    @property
    def name(self) -> str:
        return self._name

    @property
    def size(self) -> SizeType:
        return self._size

    @size.setter
    def size(self, value: SizeType):
        if not isinstance(value, (int, Symbol)):
            raise exceptions.FastrTypeError("Dimension.size should be an int or"
                                            " sympy.Symbol, found [{}] {}".format(type(value).__name__,
                                                                                  value))

        if isinstance(value, int) and value < 0:
            raise exceptions.FastrValueError("Dimension.size should be positive")

        self._size = value

    def update_size(self, value: SizeType):
        if self._size == value:
            # No effect
            return

        if not isinstance(value, int):
            raise exceptions.FastrTypeError("Dimension.size can only be updated by an int "
                                            " found [{}] {}".format(type(value).__name__,
                                                                    value))

        if value < 0:
            raise exceptions.FastrValueError("Dimension.size should be positive")

        if isinstance(self._size, Symbol):
            self._size = value
        else:
            self._size = max(self._size, value)

    def copy(self) -> 'Dimension':
        """
        Get a copy object of a Dimension
        """
        return Dimension(self.name, self.size)


class HasDimensions(object, metaclass=ABCMeta):
    """
    A Mixin class for any object that has a notion of dimensions and size. It
    uses the dimension property to expose the dimension name and size.
    """

    @property
    @abstractmethod
    def dimensions(self) -> Tuple['Dimension', ...]:
        """
        The dimensions has to be implemented by any subclass. It has to provide
        a tuple of Dimensions.

        :return: dimensions
        :rtype: tuple
        """

    @property
    def dimnames(self) -> Tuple[str]:
        """
        A tuple containing the dimension names of this object. All items of the
        tuple are of type str.
        """
        return tuple(x.name for x in self.dimensions)

    @property
    def size(self) -> Tuple[SizeType]:
        """
        A tuple containing the size of this object. All items of the
        tuple are of type int or Symbol.
        """
        return tuple(x.size for x in self.dimensions)

    @property
    def ndims(self) -> int:
        """
        The number of dimensions in this object
        """
        return len(self.dimensions)


class ForwardsDimensions(HasDimensions):
    """
    Class of objects that have dimensions not because they contain data with
    dimensions but forward them (optionally with changes via combine_dimensions)
    """
    @property
    @abstractmethod
    def source(self) -> HasDimensions:
        """
        The source object from which the dimensions are forwarded

        :return: the object from which the dimensions are forwarded
        :rtype: HasDimensions
        """

    @abstractmethod
    def combine_dimensions(self, dimensions) -> Tuple[Dimension, ...]:
        """
        Method to combine/manipulate the dimensions

        :param dimensions: the input dimensions from the source
        :return: dimensions manipulated for this object
        :rtype: tuple of dimensions
        """

    @property
    def dimensions(self) -> Tuple[Dimension, ...]:
        """
        The dimensions of the object based on the forwarding
        """
        return self.combine_dimensions(self.source.dimensions)
